# distutils: language = c++
#cython: language_level=3, boundscheck=False

"""PyPtxt. Plaintext of Pyfhel, Python For Homomorphic Encryption Libraries.
"""
# -------------------------------- IMPORTS ------------------------------------
# Encoding types: 0-UNDEFINED, 1-INTEGER, 2-FRACTIONAL, 3-BATCH
from Pyfhel.util import ENCODING_t

# Dereferencing pointers in Cython in a secure way
from cython.operator cimport dereference as deref

# ----------------------------- IMPLEMENTATION --------------------------------
cdef class PyPtxt:
    """Plaintext of Pyfhel. Contains a value/vector of unencrypted ints/doubles.

    This class references SEAL, PALISADE and HElib plaintexts, using the one 
    corresponding to the backend selected in Pyfhel (SEAL by default).

    Attributes:
        other_ptxt (PyPtxt, optional): Other PyPtxt to deep copy
    
    """
    
    def __cinit__(self, 
                  PyPtxt copy_ptxt=None,
                  Pyfhel pyfhel=None,
                  fileName=None,
                  encoding=None):
        if (copy_ptxt): # If there is a PyPtxt to copy, override all arguments and copy
            self._ptr_ptxt = new Plaintext(deref(copy_ptxt._ptr_ptxt))
            self._encoding = copy_ptxt._encoding
            if (copy_ptxt._pyfhel):
                self._pyfhel = copy_ptxt._pyfhel
        else:
            self._ptr_ptxt = new Plaintext()  
            if fileName:
                if not encoding:
                    raise TypeError("<Pyfhel ERROR> PyPtxt initialization with loading requires valid encoding")    
                self.from_file(fileName, encoding)
            else:
                self._encoding = to_ENCODING_t(encoding) if encoding else ENCODING_T.UNDEFINED
            if (pyfhel):
                self._pyfhel = pyfhel
                
    def __dealloc__(self):
        if self._ptr_ptxt != NULL:
            del self._ptr_ptxt
            
    @property
    def _encoding(self):
        """returns the encoding type"""
        return ENCODING_t(self._encoding)
    
    @_encoding.setter
    def _encoding(self, new_encoding):
        """Sets Encoding type: 0-UNDEFINED, 1-INTEGER, 2-FRACTIONAL, 3-BATCH""" 
        if not isinstance(new_encoding, ENCODING_t):
            raise TypeError("<Pyfhel ERROR> Encoding type of PyPtxt must be ENCODING_t")        
        self._encoding = new_encoding.value
        
    @_encoding.deleter
    def _encoding(self):
        """Sets Encoding to 1-UNDEFINED""" 
        self._encoding = ENCODING_t.UNDEFINED.value
              
        
    @property
    def _pyfhel(self):
        """A pyfhel instance, used for operations"""
        return self._pyfhel

    @_pyfhel.setter
    def _pyfhel(self, new_pyfhel):
        """Sets the pyfhel instance, used for operations""" 
        if not isinstance(new_pyfhel, Pyfhel):
            raise TypeError("<Pyfhel ERROR> new_pyfhel needs to be a Pyfhel class object")       
        self._pyfhel = new_pyfhel 
        
        
    cpdef bool is_zero(self) except +:
        """bool: Flag to quickly check if it is empty"""
        return self._ptr_ptxt.is_zero()

    cpdef string to_poly_string(self) except +:
        """string: Polynomial representation of the plaintext"""
        return self._ptr_ptxt.to_string()
    
    
    # =========================================================================
    # ================================== I/O ==================================
    # =========================================================================
    cpdef void to_file(self, fileName) except +:
        """to_file(Path fileName)
        
        Alias of `save` with input sanitizing.
        """
        self.save(_to_valid_file_str(fileName))

    cpdef void save(self, str fileName) except +:
        """Save the plaintext into a file.

        Args:
            fileName: (:obj:`str`) File where the plaintext will be stored.

        """
        cdef ofstream* outputter
        cdef string bFileName = fileName.encode('utf8')
        outputter = new ofstream(bFileName, binary)
        try:
            self._ptr_ptxt.save(deref(outputter))
        finally:
            del outputter

    cpdef bytes to_bytes(self) except +:
        """to_bytes()

        Serialize the plaintext into a binary/bytes string.

        Return:
            * bytes: serialized plaintext
        """
        cdef ostringstream outputter
        self._ptr_ptxt.save(outputter)
        return outputter.str()

    cpdef void from_file(self, fileName, encoding) except +:
        """from_file(str fileName)
        
        Alias of `load` with input sanitizer.
        """
        self.load(_to_valid_file_str(fileName, check=True), encoding)

    cpdef void load(self, str fileName, encoding) except +:
        """load(self, str fileName)
        
        Load the plaintext from a file.

        Args:
            fileName: (:obj:`str`) File where the plaintext is retrieved from.
            encoding: (:obj: `str`) String or type describing the encoding:
                'int' or int for IntegerEncoding (default),
                'float'/'fractional'/'double' or float for FractionalEncoding,
                'array'/'batch'/'matrix' or list for BatchEncoding

        """
        cdef ifstream* inputter
        cdef string bFileName = fileName.encode('utf8')
        inputter = new ifstream(bFileName,binary)
        try:
            self._ptr_ptxt.load(deref(inputter))
        finally:
            del inputter
        self._encoding = to_ENCODING_t(encoding).value

    cpdef void from_bytes(self, bytes content, encoding) except +:
        """from_bytes(bytes content)

        Recover the serialized plaintext from a binary/bytes string.

        Args:
            content: (:obj:`bytes`) Python bytes object containing the PyPtxt.
            encoding: (:obj: `str`) String or type describing the encoding:
                'int' or int for IntegerEncoding (default),
                'float'/'fractional'/'double' or float for FractionalEncoding,
                'array'/'batch'/'matrix' or list for BatchEncoding
        """
        cdef stringstream inputter
        inputter.write(content,len(content))
        self._ptr_ptxt.load(inputter)
        self._encoding = to_ENCODING_t(encoding).value



    # =========================================================================
    # ============================ ENCR/DECR/CMP ==============================
    # =========================================================================

    def __int__(self):
        if (self._encoding != ENCODING_T.INTEGER):
            raise RuntimeError("<Pyfhel ERROR> wrong PyCtxt encoding (not INTEGER)")
        return self._pyfhel.decodeInt(self)

    def __float__(self):
        if (self._encoding != ENCODING_T.FRACTIONAL):
            raise RuntimeError("<Pyfhel ERROR> wrong PyCtxt encoding (not FRACTIONAL)")
        return self._pyfhel.decodeFrac(self)
    
    def __repr__(self):
        return "<Pyfhel Ciphertext, encoding={}, poly={}>".format(
                ENCODING_t(self._encoding).name,
                str(self.to_poly_string())[:25]+'...' if len(str(self.to_poly_string()))>25 else '')

    def encode(self, value):
        self._pyfhel.encode(value, self)
    
    def decode(self):
        self._pyfhel.decode(self)


# Encrypting Demo for Pyfhel, covering the different ways of encrypting
#   and decrypting.

from Pyfhel import Pyfhel, PyPtxt, PyCtxt

print("==============================================================")
print("===================== Pyfhel ENCRYPTING ======================")
print("==============================================================")


print("1. Creating Context and KeyGen in a Pyfhel Object ")
HE = Pyfhel()                                       # Creating empty Pyfhel object
HE.contextGen(p=65537, m=2048, base=3, flagBatching=True)   # Generating context. 
# The values of p and m are chosen to enable batching (see Demo_Batching_SIMD.py)
HE.keyGen()                                         # Key Generation.

print("2. Encrypting integers with encryptInt")
integer1 = 94
integer2 = -235
ctxt_i1 = HE.encryptInt(integer1)   # Encrypting integer1 in a new PyCtxt with encryptInt
ctxt_i2 = PyCtxt()                  # Empty ciphertexts have no encoding type.
print("    Empty created ctxt_i2: ",   str(ctxt_i2))
HE.encryptInt(integer2, ctxt_i2)    # Encrypting integer2 in an existing PyCtxt
print("    int ",integer1,'-> ctxt_i1 ', str(ctxt_i1))
print("    int ",integer2,'-> ctxt_i2 ', str(ctxt_i2))

print("3. Encrypting floating point values with encryptFrac")
float1 = 3.5
float2 = -7.8
ctxt_f1 = HE.encryptFrac(float1)     # Encrypting float1 in a new PyCtxt with encryptFrac
ctxt_f2 = PyCtxt()
HE.encryptFrac(float2, ctxt_f2)     # Encrypting float2 in an existing PyCtxt
print("    float ",float1,'-> ctxt_f1 ', str(ctxt_f1))
print("    float ",float2,'-> ctxt_f2 ', str(ctxt_f2))

print("4. Encrypting lists of integers using batching with encryptBatch")
vector1 = [ 1, 2, 3, 4, 5, 6]
vector2 = [-2, 3,-4,-3, 2,-1]
ctxt_b1 = HE.encryptBatch(vector1)  # Encrypting vector1 in a new PyCtxt with encryptBatch
ctxt_b2 = PyCtxt()                 
HE.encryptBatch(vector2, ctxt_b2)    # Encrypting vector2 in an existing PyCtxt
print("    list ",vector1,'-> ctxt_b1 ', str(ctxt_b1))
print("    list ",vector2,'-> ctxt_b2 ', str(ctxt_b2))

print("5. Encrypting numpy 1D integer vectors using batching encryptArray")
import numpy as np
array1 = np.array([-6, -5, -4, -3, -2, -1],dtype=np.int64)
array2 = np.array([12, 15, 18, 21, 24, 27],dtype=np.int64)
ctxt_a1 = HE.encryptArray(array1)   # Encrypting array1 in a new PyCtxt with encryptArray
ctxt_a2 = PyCtxt()                
HE.encryptArray(array2, ctxt_a2)    # Encrypting vector2 in an existing PyCtxt
print("    array ",array1,'-> ctxt_a1 ', str(ctxt_a1))
print("    array ",array2,'-> ctxt_a2 ', str(ctxt_a2))

print("6. Operations in the encrypted domain")
ctxt_iadd = ctxt_i1 + ctxt_i2     # `ctxt_i1 += ctxt_i2` for quicker inplace operation
ctxt_isub = ctxt_i1 - ctxt_i2     # `ctxt_i1 -= ctxt_i2` for quicker inplace operation
ctxt_imul = ctxt_i1 * ctxt_i2     # `ctxt_i1 *= ctxt_i2` for quicker inplace operation

ctxt_fadd = ctxt_f1 + ctxt_f2     # `ctxt_f1 += ctxt_f2` for quicker inplace operation
ctxt_fsub = ctxt_f1 - ctxt_f2     # `ctxt_f1 -= ctxt_f2` for quicker inplace operation
ctxt_fmul = ctxt_f1 * ctxt_f2     # `ctxt_f1 *= ctxt_f2` for quicker inplace operation

ctxt_badd = ctxt_b1 + ctxt_b2     # `ctxt_b1 += ctxt_b2` for quicker inplace operation
ctxt_bsub = ctxt_b1 - ctxt_b2     # `ctxt_b1 -= ctxt_b2` for quicker inplace operation
ctxt_bmul = ctxt_b1 * ctxt_b2     # `ctxt_b1 *= ctxt_b2` for quicker inplace operation

ctxt_aadd = ctxt_a1 + ctxt_a2     # `ctxt_a1 += ctxt_a2` for quicker inplace operation
ctxt_asub = ctxt_a1 - ctxt_a2     # `ctxt_a1 -= ctxt_a2` for quicker inplace operation
ctxt_amul = ctxt_a1 * ctxt_a2     # `ctxt_a1 *= ctxt_a2` for quicker inplace operation

ctxt_abadd = ctxt_a1 + ctxt_b2    # `ctxt_a1 += ctxt_b2` for quicker inplace operation
ctxt_absub = ctxt_a1 - ctxt_b2    # `ctxt_a1 -= ctxt_b2` for quicker inplace operation
ctxt_abmul = ctxt_a1 * ctxt_b2    # `ctxt_a1 *= ctxt_b2` for quicker inplace operation


print("    PyCtxt Ciphertexts can only be operated with each other if:")
print("    I.  They have the same encoding type.")
print("        int                           - ENCODING_t.INTEGER")
print("        float                         - ENCODING_t.FRACTIONAL")
print("        list[int] and np.array([int]) - ENCODING_t.BATCH")
print("    II. They were encrypted with the same public key and context")


print("7. Decrypting result:")
res_iSum = HE.decryptInt(ctxt_iadd) # Decryption must use the corresponding decrypting
res_iSub = HE.decryptInt(ctxt_isub) #  function, depending on the encoding type
res_iMul = HE.decryptInt(ctxt_imul)

res_fSum = HE.decryptFrac(ctxt_fadd) # Decryption must use the corresponding decrypting
res_fSub = HE.decryptFrac(ctxt_fsub) #  function, depending on the encoding type
res_fMul = HE.decryptFrac(ctxt_fmul)

res_bSum = HE.decryptBatch(ctxt_badd) # Decryption must use the corresponding decrypting
res_bSub = HE.decryptBatch(ctxt_bsub) #  function, depending on the encoding type
res_bMul = HE.decryptBatch(ctxt_bmul)

res_aSum = HE.decryptArray(ctxt_aadd) # Decryption must use the corresponding decrypting
res_aSub = HE.decryptArray(ctxt_asub) #  function, depending on the encoding type
res_aMul = HE.decryptArray(ctxt_amul)

res_abSum = HE.decryptBatch(ctxt_abadd) # Decryption must use the corresponding decrypting
res_abSub = HE.decryptBatch(ctxt_absub) #  function, depending on the encoding type
res_abMul = HE.decryptArray(ctxt_abmul)

print("    Integer (encryptInt, ENCODING_t.INTEGER, decryptInt)")
print("     addition:       decrypt(ctxt_i1 + ctxt_i2) =  ", res_iSum)
print("     substraction:   decrypt(ctxt_i1 - ctxt_i2) =  ", res_iSub)
print("     multiplication: decrypt(ctxt_i1 + ctxt_i2) =  ", res_iMul)

print("    Float (encryptFrac, ENCODING_t.FRACTIONAL, decryptFrac)")
print("     addition:       decrypt(ctxt_f1 + ctxt_f2) =  ", res_fSum)
print("     substraction:   decrypt(ctxt_f1 - ctxt_f2) =  ", res_fSub)
print("     multiplication: decrypt(ctxt_f1 + ctxt_f2) =  ", res_fMul)

print("    Batched list (encryptBatch, ENCODING_t.BATCH, decryptBatch)")
print("     addition:       decrypt(ctxt_b1 + ctxt_b2) =  ", res_bSum[:len(vector1)])
print("     substraction:   decrypt(ctxt_b1 - ctxt_b2) =  ", res_bSub[:len(vector1)])
print("     multiplication: decrypt(ctxt_b1 + ctxt_b2) =  ", res_bMul[:len(vector1)])

print("    NumPy 1D vector (encryptArray, ENCODING_t.BATCH, decryptArray)")
print("     addition:       decrypt(ctxt_a1 + ctxt_a2) =  ", np.array(res_aSum[:len(array1)]))
print("     substraction:   decrypt(ctxt_a1 - ctxt_a2) =  ", np.array(res_aSub[:len(array1)]))
print("     multiplication: decrypt(ctxt_a1 + ctxt_a2) =  ", np.array(res_aMul[:len(array1)]))

print("    List & np 1D vec (..., ENCODING_t.BATCH, ...)")
print("     addition:       decrypt(ctxt_a1 + ctxt_b2) =  ", res_abSum[:len(vector1)])
print("     substraction:   decrypt(ctxt_a1 - ctxt_b2) =  ", res_abSub[:len(vector1)])
print("     multiplication: decrypt(ctxt_a1 + ctxt_b2) =  ", np.array(res_abMul[:len(vector1)]))


print("8. EXTRA: you can encrypt/decrypt function to guess type")
print("    HE.encrypt(integer1|float1|vector1|array1)")
print("       delegates into encryptInt|encryptFrac|EncryptBatch|envryptArray depending on type")
print("    HE.decrypt(PyCtxt)")
print("       uses recorded ENCODING_t type to correctly decrypt")
ctxt1 = HE.encrypt(integer1)
ctxt2 = HE.encrypt(float1)
ctxt3 = HE.encrypt(array1)

integer1 = HE.decrypt(ctxt1)
float1 =   HE.decrypt(ctxt2)
array1 =   HE.decrypt(ctxt3)



# distutils: language = c++
#cython: language_level=3, boundscheck=False
# -------------------------------- CIMPORTS ------------------------------------
from libcpp.string cimport string

# ---------------------------- CYTHON DECLARATION ------------------------------

cdef extern from "<iostream>" namespace "std":
    cdef cppclass ostream:
        ostream& write(const char*, int) except +
    cdef cppclass istream:
        istream& read(char*, int) except +

cdef extern from "<iostream>" namespace "std::ios_base":
    cdef cppclass open_mode:
        pass
    cdef open_mode binary

cdef extern from "<fstream>" namespace "std":
    cdef cppclass ofstream(ostream):
        # constructors
        ofstream()except +
        ofstream(const char*) except +
        ofstream(const char*, open_mode) except+
        ofstream(const string&) except +
        ofstream(const string&, open_mode) except+
        void open(const char*) except +
        void open(const string&) except +
        void close() except +
    cdef cppclass ifstream(istream):
        # constructors
        ifstream()except +
        ifstream(const char*) except +
        ifstream(const char*, open_mode) except+
        ifstream(const string&) except +
        ifstream(const string&, open_mode) except+
        void open(const char&) except +
        void open(const string&) except +
        void close() except +

cdef extern from "<sstream>" namespace "std":
    cdef cppclass stringstream(istream):
        istringstream() except +
        istringstream(const char*) except +
        istringstream(const string&) except +
        stringstream& write(const char*, int) except +
    cdef cppclass ostringstream(ostream):
        ostringstream() except +
        string str() except +



# ContextParameters shows how several parameters affect performance.
#  TODO: Finish for all parameters. Show possible values of each.

import time

from Pyfhel import Pyfhel, PyPtxt, PyCtxt
# Pyfhel class contains most of the functions.
# PyPtxt is the plaintext class
# PyCtxt is the ciphertext class


print("==============================================================")
print("================== Pyfhel CONTEXT PARAMETERS =================")
print("==============================================================")


print("1. p (long): Plaintext modulus. All operations are modulo p. ")
HE = Pyfhel()           # Creating empty Pyfhel object
HE.contextGen(p=65537)  # Generating context. The value of p is important.
                        #  There are many configurable parameters on this step
HE.keyGen()             # Key Generation.


print("2. m (long=2048): Polynomial coefficient modulus.")
print("   Higher allows more encrypted operations. In batch mode it is the number of integers per ciphertext.")

def time_demo_helloworld(integer1=127, integer2=-2):
    start = time.time()
    ctxt1 = HE.encryptInt(integer1) # Encryption makes use of the public key
    ctxt2 = HE.encryptInt(integer2) # For integers, encryptInt function is used.
    ctxtSum = ctxt1 + ctxt2         # `ctxt1 += ctxt2` for quicker inplace operation
    ctxtSub = ctxt1 - ctxt2         # `ctxt1 -= ctxt2` for quicker inplace operation
    ctxtMul = ctxt1 * ctxt2         # `ctxt1 *= ctxt2` for quicker inplace operation
    resSum = HE.decryptInt(ctxtSum)
    resSub = HE.decryptInt(ctxtSub) 
    resMul = HE.decryptInt(ctxtMul)
    end = time.time()
    print(f"Elapsed time: {end - start}")
    results = (resSum==125, resSub==129, resMul==-254)
    return end - start, results


# Fractional / float Demo for Pyfhel, covering the different ways of encrypting
#   and decrypting.

from Pyfhel import Pyfhel, PyPtxt, PyCtxt

print("==============================================================")
print("============ Pyfhel MultDepth and Relineraization ============")
print("==============================================================")
print(" This demo focuses on ciphertext size and noise budget")

print("1. Creating Context and KeyGen in a Pyfhel Object. Using 64 ")
print("     bits for integer part and 32 bits for decimal part.")
HE = Pyfhel()           # Creating empty Pyfhel object
HE.contextGen(p=65537, base=2, intDigits=64, fracDigits = 32) 
                        # Generating context. The value of p is important.
                        #  There are many configurable parameters on this step
                        #  More info in Demo_ContextParameters.py, and
                        #  in the docs of the function (link to docs in README)
HE.keyGen()             # Key Generation.
print(HE)

print("2. Encrypting Fractionals")
float1 = -2.3
float2 = 3.4
ctxt1 = HE.encryptFrac(float1) # Encryption makes use of the public key
ctxt2 = HE.encryptFrac(float2) # For integers, encryptInt function is used.
print("    int ",float1,'-> ctxt1 ', type(ctxt1))
print("    int ",float2,'-> ctxt2 ', type(ctxt2))
print(ctxt1)
print(ctxt2)


print("3. Operating with encrypted fractionals multiple times")
print("     ctxtMul1 = ctxt1 * ctxt2")
ctxtMul1 = ctxt1 * ctxt2         # `ctxt1 *= ctxt2` for quicker inplace operation
print(ctxtMul1)
print("   Notice the increase in size: size(ctxtMul1) = size(ctxt1) + size(ctxt2) - 1")
print("   This rule applies for all multiplications.")
print("   This size increase makes further operations slower, and requires more memory")


print("4. Relinearization reduces the size of a ciphertext back to 2, thus it is very convenient for deep multiplications")
print("    First we generate relinearization keys.")
print("    Make sure the bitcount is higher than the size of your ciphertexts")
HE.relinKeyGen(1, 1)

ctxtMul1 = ~ ctxtMul1 # Equivalent to HE.relinearize(ctxtMul)
                      # and equivalent to  ~ctxtMul1 (without assignment)
print(ctxtMul1)
print("   Good! size increase is solved with relinearization.")


print("5. Noise budget: You can see ciphertexts as noisy signals. Every multiplication will drastically increase the noise, thus reducing the noise budget.")
print("   Notice the decrease in noise budget in the previous multiplication")
print("   If the noise budget reaches zero, we won't be able to decrypt correctly!")
print("ctxtMulsq = ctxtMult1**2")
ctxtMulsq = ctxtMul1**2
print(ctxtMulsq)
print("     decrypt(ctxtMul1) =  ", HE.decryptFrac(ctxtMul1))
print(f"      expected: {float1*float2}")
print("     decrypt(ctxtMulsq) =  ", HE.decryptFrac(ctxtMulsq))
print(f"      expected: {(float1*float2)**2}")

print("6. I need more noise budget! You have several options:")
print("  - Use better context parameters (mainly increase m, reduce p, reduce intDigits and fracDigits")
print("    -> Try this demo changing p to be 63 (enough to hold the result value) -> precision lowers, but operations are faster and noise budget decreases slower")
print("    -> Try this demo with m=8192 (more multiplications are allowed, but operations are slower")
print("  - Decrypt and encrypt again.")
print("  - Use bootstrapping: Sadly it is not available in SEAL, and even if it were, it would be extremely inefficient. It is always better to resort to the first option")
HE



# Encoding Demo for Pyfhel, covering the different ways of encodeing
#   and decodeing.

from Pyfhel import Pyfhel, PyPtxt, PyPtxt

print("==============================================================")
print("====================== Pyfhel ENCODING =======================")
print("==============================================================")


print("1. Creating Context and KeyGen in a Pyfhel Object ")
HE = Pyfhel()                                       # Creating empty Pyfhel object
HE.contextGen(p=65537, m=1024, flagBatching=True)   # Generating context. 
# The values of p and m are chosen to enable batching (see Demo_Batching_SIMD.py)
HE.keyGen()                                         # Key Generation.

print("2. Encoding integers with encodeInt")
integer1 = 45
integer2 = -32
ptxt_i1 = HE.encodeInt(integer1)   # Encoding integer1 in a new PyPtxt
ptxt_i2 = PyPtxt()                 # Empty plaintexts have no encoding type.
print("    Empty created ptxt_i2: ",   str(ptxt_i2))
HE.encodeInt(integer2, ptxt_i2)    # Encoding integer2 in an existing PyPtxt
print("    int ",integer1,'-> ptxt_i1 ', str(ptxt_i1))
print("    int ",integer2,'-> ptxt_i2 ', str(ptxt_i2))

print("3. Encoding floating point values with encodeFrac")
float1 = 3.5
float2 = -7.8
ptxt_f1 = HE.encodeFrac(float1)     # Encoding float1 in a new PyPtxt with encodeFrac
ptxt_f2 = PyPtxt()
HE.encodeFrac(float2, ptxt_f2)     # Encoding float2 in an existing PyPtxt
print("    float ",float1,'-> ptxt_f1 ', str(ptxt_f1))
print("    float ",float2,'-> ptxt_f2 ', str(ptxt_f2))

print("4. Encoding lists of integers using batching with encodeBatch")
vector1 = [ 1, 2, 3, 4, 5, 6]
vector2 = [-2, 3,-4,-3, 2,-1]
ptxt_b1 = HE.encodeBatch(vector1)  # Encoding vector1 in a new PyPtxt with encodeBatch
ptxt_b2 = PyPtxt()                 
HE.encodeBatch(vector2, ptxt_f2)    # Encoding vector2 in an existing PyPtxt
print("    list ",vector1,'-> ptxt_b1 ', str(ptxt_b1))
print("    list ",vector2,'-> ptxt_b2 ', str(ptxt_b2))

print("5. Encoding numpy 1D integer vectors using batching encodeArray")
import numpy as np
array1 = np.array([-6, -5, -4, -3, -2, -1],dtype=np.int64)
array2 = np.array([12, 15, 18, 21, 24, 27],dtype=np.int64)
ptxt_a1 = HE.encodeArray(array1)   # Encoding array1 in a new PyPtxt with encodeArray
ptxt_a2 = PyPtxt()                
HE.encodeArray(array2, ptxt_a2)    # Encoding vector2 in an existing PyPtxt
print("    array ",array1,'-> ptxt_a1 ', str(ptxt_a1))
print("    array ",array2,'-> ptxt_a2 ', str(ptxt_a2))

print("6. Encrypting/Decrypting all plaintexts using encryptPtxt/decryptPtxt")
print("      HE.encrypt could also do, detecting PyPtxt as a valid type")
print("      Just remember to operate only with ciphertexts with same encoding")
print("     ATTENTION: HE.decrypt will directly decrypt AND decode if used with decode_value=True! ")
ctxt1 = HE.encryptPtxt(ptxt_i1)
ctxt2 = HE.encrypt(ptxt_f1)
ctxt3 = HE.encryptPtxt(ptxt_b1)
ctxt4 = HE.encrypt(ptxt_a1)

integer1 = HE.decrypt(ctxt1)
float1 = HE.decrypt(ctxt2)
vector1 = HE.decrypt(ctxt3)

print("7. Decoding values")
res_i1 = HE.decodeInt(ptxt_i1) # Decoding must use the corresponding decodeing
res_f1 = HE.decodeFrac(ptxt_f1) # Decoding must use the corresponding decodeing
res_b1 = HE.decodeBatch(ptxt_b1) # Decoding must use the corresponding decodeing
res_a1 = HE.decodeArray(ptxt_a1) # Decoding must use the corresponding decodeing

print("    Integer (encodeInt, ENCODING_t.INTEGER, decodeInt)")
print("      decode(ptxt_i1) =  ", res_i1)

print("    Float (encodeFrac, ENCODING_t.FRACTIONAL, decodeFrac)")
print("     decode(ptxt_f1) =  ", res_f1)

print("    Batched list (encodeBatch, ENCODING_t.BATCH, decodeBatch)")
print("     decode(ptxt_b1) =  ", res_b1[:len(vector2)])

print("    NumPy 1D vector (encodeArray, ENCODING_t.BATCH, decodeArray)")
print("     decode(ptxt_a1) =  ", res_a1[:len(array2)])

print("8. EXTRA: you can use encode/decode function to guess type")
print("    HE.encode(integer1|float1|vector1|array1)")
print("       delegates into encodeInt|encodeFrac|encodeBatch|encodeArray depending on type")
print("    HE.decode(PyPtxt)")
print("       uses recorded ENCODING_t type to correctly decode")
ptxt1 = HE.encode(integer2)
ptxt2 = HE.encode(float2)
ptxt3 = HE.encode(array2)

integer2 = HE.decode(ptxt1)
float2 =   HE.decode(ptxt2)
array2 =   HE.decode(ptxt3)



# distutils: language = c++
#cython: language_level=3, boundscheck=False

# -------------------------------- CIMPORTS ------------------------------------
# Import from Cython libs required C/C++ types for the Afhel API
from libcpp.string cimport string
from libcpp cimport bool

# Used for all kinds of operations. Includes utility functions
from Pyfhel.Pyfhel cimport *

# Import our own wrapper for iostream classes, used for I/O ops
from Pyfhel.iostream cimport ifstream, ofstream, ostringstream, stringstream, binary

# Import Ciphertext class, original for SEAL
from Pyfhel.Afhel cimport Ciphertext

# Encoding types: 0-UNDEFINED, 1-INTEGER, 2-FRACTIONAL, 3-BATCH
from Pyfhel.util cimport ENCODING_T

# ---------------------------- CYTHON DECLARATION ------------------------------
cdef class PyCtxt:
    cdef Ciphertext* _ptr_ctxt
    cdef Pyfhel _pyfhel
    cdef ENCODING_T _encoding
    cpdef int size_capacity(self) except +
    cpdef int size(self) except +
    cpdef void to_file(self, fileName) except +
    cpdef void from_file(self, fileName, encoding) except +
    cpdef void save(self, str fileName) except +
    cpdef void load(self, str fileName, encoding) except +
    cpdef bytes to_bytes(self) except +
    cpdef void from_bytes(self, bytes content, encoding) except +


# distutils: language = c++
#cython: language_level=3, boundscheck=False

# -------------------------------- CIMPORTS -----------------------------------
# import both numpy and the Cython declarations for numpy
cimport numpy as np

# Import from Cython libs required C/C++ types for the Afhel API
from libcpp.vector cimport vector
from libcpp cimport bool
from numpy cimport int64_t, uint64_t

# Import our own wrapper for iostream classes, used for I/O ops
from Pyfhel.iostream cimport istream, ostream, ifstream, ofstream, binary

from Pyfhel.Afhel cimport Plaintext
from Pyfhel.Afhel cimport Ciphertext
from Pyfhel.Afhel cimport Afseal

# Import the Cython Plaintext and Cyphertext classes
from Pyfhel.PyPtxt cimport PyPtxt
from Pyfhel.PyCtxt cimport PyCtxt

# Encoding types: 0-UNDEFINED, 1-INTEGER, 2-FRACTIONAL, 3-BATCH
from Pyfhel.util cimport ENCODING_T


# ---------------------------- CYTHON DECLARATION ------------------------------
cdef class Pyfhel:
    cdef Afseal* afseal           # The C++ methods are accessed via a pointer
    
    # =========================== CRYPTOGRAPHY =================================
    cpdef contextGen(self, long p, long m=*, bool flagBatching=*, long base=*,
                     long sec=*, int intDigits=*, int fracDigits=*) except +
    cpdef void keyGen(self) except +
    cpdef void rotateKeyGen(self, int bitCount) except +
    cpdef void relinKeyGen(self, int bitCount, int size) except +
    
    cpdef PyCtxt encryptInt(self, int64_t value, PyCtxt ctxt=*) except +
    cpdef PyCtxt encryptFrac(self, double value, PyCtxt ctxt=*) except +
    cpdef PyCtxt encryptBatch(self, vector[int64_t] vec, PyCtxt ctxt=*) except +
    cpdef PyCtxt encryptArray(self, int64_t[::1] arr, PyCtxt ctxt=*) except +
    cpdef PyCtxt encryptPtxt(self, PyPtxt ptxt, PyCtxt ctxt=*) except +
    
    cpdef int64_t decryptInt(self, PyCtxt ctxt) except +
    cpdef double decryptFrac(self, PyCtxt ctxt) except +
    cpdef vector[int64_t] decryptBatch(self, PyCtxt ctxt) except +
    cpdef int64_t[::1] decryptArray(self, PyCtxt ctxt) except +
    cpdef PyPtxt decryptPtxt(self, PyCtxt ctxt, PyPtxt ptxt=*) except +
    
    cpdef void relinearize(self, PyCtxt ctxt) except +
    
    cpdef int noiseLevel(self, PyCtxt ctxt) except +
    
    # ============================= ENCODING ===================================
    cpdef PyPtxt encodeInt(self, int64_t value, PyPtxt ptxt=*) except +
    cpdef PyPtxt encodeFrac(self, double value, PyPtxt ptxt=*) except +
    cpdef PyPtxt encodeBatch(self, vector[int64_t] vec, PyPtxt ptxt=*) except +
    cpdef PyPtxt encodeArray(self, int64_t[::1] arr, PyPtxt ptxt=*) except +
    
    cpdef int64_t decodeInt(self, PyPtxt ptxt) except +
    cpdef double decodeFrac(self, PyPtxt ptxt) except +
    cpdef vector[int64_t] decodeBatch(self, PyPtxt ptxt) except +
    cpdef int64_t[::1] decodeArray(self, PyPtxt ptxt) except +
    
    # ============================ OPERATIONS ==================================
    cpdef PyCtxt square(self, PyCtxt ctxt, bool in_new_ctxt=*) except +
    cpdef PyCtxt negate(self, PyCtxt ctxt, bool in_new_ctxt=*) except +
    cpdef PyCtxt add(self, PyCtxt ctxt, PyCtxt ctxt_other, bool in_new_ctxt=*) except +
    cpdef PyCtxt add_plain(self, PyCtxt ctxt, PyPtxt ptxt, bool in_new_ctxt=*) except +
    cpdef PyCtxt sub(self, PyCtxt ctxt, PyCtxt ctxt_other, bool in_new_ctxt=*) except +
    cpdef PyCtxt sub_plain(self, PyCtxt ctxt, PyPtxt ptxt, bool in_new_ctxt=*) except +
    cpdef PyCtxt multiply(self, PyCtxt ctxt, PyCtxt ctxt_other, bool in_new_ctxt=*) except +
    cpdef PyCtxt multiply_plain(self, PyCtxt ctxt, PyPtxt ptxt, bool in_new_ctxt=*) except +
    cpdef PyCtxt rotate(self, PyCtxt ctxt, int k, bool in_new_ctxt=*) except +
    cpdef PyCtxt power(self, PyCtxt ctxt, uint64_t expon, bool in_new_ctxt=*) except +
    cpdef PyCtxt polyEval(self, PyCtxt ctxt, vector[int64_t] coeffPoly, bool in_new_ctxt=*) except +
    cpdef PyCtxt polyEval_double (self, PyCtxt ctxt, vector[double] coeffPoly, bool in_new_ctxt=*) except +

    
    # ================================ I/O =====================================
    cpdef bool saveContext(self, fileName) except +
    cpdef bool restoreContext(self, fileName) except +

    cpdef bool savepublicKey(self, fileName) except +
    cpdef bool restorepublicKey(self, fileName) except +

    cpdef bool savesecretKey(self, fileName) except +
    cpdef bool restoresecretKey(self, fileName) except +

    cpdef bool saverelinKey(self, fileName) except +
    cpdef bool restorerelinKey(self, fileName) except +

    cpdef bool saverotateKey(self, fileName) except +
    cpdef bool restorerotateKey(self, fileName) except +

    
    # ============================== AUXILIARY =================================
    cpdef bool batchEnabled(self) except +
    cpdef long relinBitCount(self) except +

    # GETTERS
    cpdef int getnSlots(self) except +
    cpdef int getp(self) except +
    cpdef int getm(self) except +
    cpdef int getbase(self) except +
    cpdef int getsec(self) except + 
    cpdef int getintDigits(self) except +
    cpdef int getfracDigits(self) except +
    cpdef bool getflagBatch(self) except +
    
    cpdef bool is_secretKey_empty(self) except+
    cpdef bool is_publicKey_empty(self) except+
    cpdef bool is_rotKey_empty(self) except+
    cpdef bool is_relinKey_empty(self) except+
    cpdef bool is_context_empty(self) except+


# --------------------------------- UTILS --------------------------------------
cpdef to_ENCODING_t(encoding) except +
cpdef str _to_valid_file_str(fileName, bool check=*) except +



# Usage of Pyfhel with Numpy
import numpy as np

from Pyfhel import Pyfhel, PyPtxt, PyCtxt
# Pyfhel class contains most of the functions.
# PyPtxt is the plaintext class
# PyCtxt is the ciphertext class


print("==============================================================")
print("===================== Pyfhel with Numpy ======================")
print("==============================================================")


print("1. Creating Context and KeyGen in a Pyfhel Object ")
HE = Pyfhel()           # Creating empty Pyfhel object
HE.contextGen(p=65537)  # Generating context. The value of p is important.
                        #  There are many configurable parameters on this step
                        #  More info in Demo_ContextParameters.py, and
                        #  in the docs of the function (link to docs in README)
HE.keyGen()             # Key Generation.
print(HE)

print("2. Encrypting two arrays of integers.")
print("    For this, you need to create empty arrays in numpy and assign them the cyphertexts")
array1 = np.array([1,3,5,7,9])
array2 = np.array([-2, 4, -6, 8,-10])
arr_ctxt1 = np.empty(len(array1),dtype=PyCtxt)
arr_ctxt2 = np.empty(len(array1),dtype=PyCtxt)

# Encrypting! This can be parallelized!
for i in np.arange(len(array1)):
    arr_ctxt1[i] = HE.encryptInt(array1[i])
    arr_ctxt2[i] = HE.encryptInt(array2[i])

print("    array1: ",array1,'-> ctxt1 ', type(arr_ctxt1), ', dtype:', arr_ctxt1.dtype)
print("    array2: ",array2,'-> ctxt2 ', type(arr_ctxt2), ', dtype:', arr_ctxt2.dtype)

print("3. Vectorized operations with encrypted arrays of PyCtxt")
ctxtSum = arr_ctxt1 + arr_ctxt2         # `ctxt1 += ctxt2` for quicker inplace operation
ctxtSub = arr_ctxt1 - arr_ctxt2         # `ctxt1 -= ctxt2` for quicker inplace operation
ctxtMul = arr_ctxt1 * arr_ctxt2         # `ctxt1 *= ctxt2` for quicker inplace operation


print("4. Decrypting results:")
resSum = [HE.decryptInt(ctxtSum[i]) for i in np.arange(len(ctxtSum))]
resSub = [HE.decryptInt(ctxtSub[i]) for i in np.arange(len(ctxtSub))] 
resMul = [HE.decryptInt(ctxtMul[i]) for i in np.arange(len(ctxtMul))]
print("     addition:       decrypt(ctxt1 + ctxt2) =  ", resSum)
print("     substraction:   decrypt(ctxt1 - ctxt2) =  ", resSub)
print("     multiplication: decrypt(ctxt1 + ctxt2) =  ", resMul)



# HelloWorld Demo for Pyfhel, showing the simplest use of the library by
#   encrypting two integers, operating with them (+,-,*) and decrypting
#   the results. The code is heavily commented.


from Pyfhel import Pyfhel, PyPtxt, PyCtxt
# Pyfhel class contains most of the functions.
# PyPtxt is the plaintext class
# PyCtxt is the ciphertext class


print("==============================================================")
print("===================== Pyfhel HELLO WORLD =====================")
print("==============================================================")


print("1. Creating Context and KeyGen in a Pyfhel Object. ")
HE = Pyfhel()           # Creating empty Pyfhel object
HE.contextGen(p=65537)  # Generating context. The value of p is important.
                        # Generating context. The value of p is important.
                        #  There are many configurable parameters on this step
                        #  More info in Demo_ContextParameters.py, and
                        #  in the docs of the function (link to docs in README)
HE.keyGen()             # Key Generation.
print(HE)

print("2. Encrypting integers")
integer1 = 127
integer2 = -2
ctxt1 = HE.encryptInt(integer1) # Encryption makes use of the public key
ctxt2 = HE.encryptInt(integer2) # For integers, encryptInt function is used.
print("    int ",integer1,'-> ctxt1 ', type(ctxt1), str(ctxt1))
print("    int ",integer2,'-> ctxt2 ', type(ctxt2), str(ctxt2))
print(ctxt1)
print(ctxt2)

print("3. Operating with encrypted integers")
ctxtSum = ctxt1 + ctxt2         # `ctxt1 += ctxt2` for quicker inplace operation
ctxtSub = ctxt1 - ctxt2         # `ctxt1 -= ctxt2` for quicker inplace operation
ctxtMul = ctxt1 * ctxt2         # `ctxt1 *= ctxt2` for quicker inplace operation
print(f"Sum: {ctxtSum}")
print(f"Sub: {ctxtSub}")
print(f"Mult:{ctxtMul}")

print("4. Decrypting result:")
resSum = HE.decryptInt(ctxtSum) # Decryption must use the corresponding function
                                #  decryptInt.
resSub = HE.decryptInt(ctxtSub) 
resMul = HE.decryptInt(ctxtMul)
print("     addition:       decrypt(ctxt1 + ctxt2) =  ", resSum)
print("     substraction:   decrypt(ctxt1 - ctxt2) =  ", resSub)
print("     multiplication: decrypt(ctxt1 + ctxt2) =  ", resMul)





from Pyfhel import Pyfhel
from PyPtxt import PyPtxt
from PyCtxt import PyCtxt
import numpy as np
import time
import math

HE = Pyfhel()
def runTest(fileName, testID, p=257, r=1, d=1, c=2, sec=80, w=64, L=10, m=-1, R=3, s=0, gens=[], ords=[]):

    KEYGEN_PARAMS={ "p":257,      "r":1,
                    "d":1,        "c":2,
                    "sec":80,     "w":64,
                    "L":10,       "m":-1,
                    "R":3,        "s":0,
                    "gens":[],    "ords":[]}
   
    KEYGEN_PARAMS['p']=p   
    KEYGEN_PARAMS['r']=r   
    KEYGEN_PARAMS['d']=d   
    KEYGEN_PARAMS['c']=c   
    KEYGEN_PARAMS['sec']=sec
    KEYGEN_PARAMS['w']=w
    KEYGEN_PARAMS['L']=L
    KEYGEN_PARAMS['m']=m
    KEYGEN_PARAMS['R']=R
    KEYGEN_PARAMS['s']=s
    KEYGEN_PARAMS['gens']=gens
    KEYGEN_PARAMS['ords']=ords

    print("Pyfhel TEST %d"%(testID))
    print(" File: "+fileName)

    # KEYGEN
    print(" KEYGEN")
    print("  Running KeyGen with params:")
    print(KEYGEN_PARAMS)
    tic_keyGen = time.time()
    HE.keyGen(KEYGEN_PARAMS)
    toc_keyGen = time.time()
    print("  KeyGen completed")
    print("  nSlots = %d"%(HE.numSlots()))

    # DATA CREATION
    v1 = [1,2,3,4,5,6,7,8,9,10]
    v2 = [2,2,2,2,2,2,2,2,2,2]
    v3 = [1,1,1,1,1,1,1,1,1,1]
    p1 = PyPtxt(v1, HE)
    p2 = PyPtxt(v2, HE)
    p3 = PyPtxt(v3, HE)

    # ENCRYPTION
    print(" ENCRYPTION")
    print("  Encrypting p1, p2, p3 into c1, c2, c3")
    tic_Encr = time.time()
    c1 = HE.encrypt(p1)
    c2 = HE.encrypt(p2)
    c3 = HE.encrypt(p3)
    toc_Encr = time.time()
    t_e = toc_Encr - tic_Encr
    print("  Encryption completed")
    
    # TIMING OPERATIONS
    print(" OPERATIONS")
    print("  c1 += c2")
    tic_Sum = time.time()
    c1 += c2
    e1 = list(np.mod(np.array(v1)+np.array(v2), math.pow(p, r)))
    toc_Sum = time.time()
    tic_Decr_Sum = time.time()
    res1 = HE.decrypt(c1)
    toc_Decr_Sum = time.time()
    t_sum = toc_Sum - tic_Sum
    t_d_sum = toc_Decr_Sum - tic_Decr_Sum
    f_sum = res1[0] == e1
    print("    Result: "+str(res1[0]))
    print("    Debug: "+str(e1))

    print("  c1 *= c2")
    c1 = HE.encrypt(p1)
    tic_Prod = time.time()
    c1 *= c2
    e2 = list(np.mod(np.array(v1) * np.array(v2), math.pow(p,r)))
    toc_Prod = time.time()
    tic_Decr_Prod = time.time()
    res2 = HE.decrypt(c1)
    toc_Decr_Prod = time.time()
    t_prod = toc_Prod - tic_Prod
    t_d_prod = toc_Decr_Prod - tic_Decr_Prod
    f_prod = res2[0] == e2
    print("    Result: "+str(res2[0]))
    print("    Debug: "+str(e2))

    print("  c1 %= c2")
    c1 = HE.encrypt(p1)
    tic_scpr = time.time()
    c1 %= c2
    e3 = np.mod(sum(np.mod((np.array(v1) * np.array(v2)), math.pow(p,r))),math.pow(p,r))
    toc_scpr = time.time()
    tic_Decr_scpr = time.time()
    res3 = HE.decrypt(c1)
    toc_Decr_scpr = time.time()
    t_scpr = toc_scpr - tic_scpr
    t_d_scpr = toc_Decr_scpr - tic_Decr_scpr
    f_scpr = res3[0][0] == e3
    print("    Result: "+str(res3[0][0]))
    print("    Debug: "+str(e3))

    # CHECKING LIMITS
    print("  Maximum number of Sums")
    t_maxSum = 0
    t_maxSd = 0
    n_maxSum = 0
    resCorrect = 1
    a = np.array(v2)
    b = np.array(v3)
    while (resCorrect):
        tic_aux = time.time()
        c2 += c3
        toc_aux = time.time()
        a = list(a + b)
        tic_ad = time.time()
        res_aux = HE.decrypt(c2)
        toc_ad = time.time()
        resCorrect = (a == res_aux[0])
        n_maxSum += 1
        t_maxSum += (toc_aux - tic_aux)
        t_maxSd += (toc_ad - tic_ad)
    n_maxSum -= 1

    print("  Maximum number of Mults")
    t_maxMult = 0
    t_maxMd = 0
    n_maxMult = 0
    resCorrect = 1
    v2 = [2,2,2,2,2,2,2,2,2,2]
    p2 = PyPtxt(v2, HE)
    c2 = HE.encrypt(p2)
    a = np.array(v2)
    while (resCorrect):
        tic_aux = time.time()
        c2 *= c3
        toc_aux = time.time()
        ab = list(a * b)
        tic_ad = time.time()
        res_aux = HE.decrypt(c2)
        toc_ad = time.time()
        resCorrect = (ab == res_aux[0])
        n_maxMult += 1
        t_maxMult += (toc_aux - tic_aux)
        t_maxMd += (toc_ad - tic_ad)
    n_maxMult -= 1


    # SAVE KEYGEN DATA ON FILE
    f = open(fileName, 'a')
    f.write("----------------- TESTCASE %d ----------------\n"%(testID))
    f.write(" KeyGen: " + str(KEYGEN_PARAMS)+"\n")
    f.write(" -  nslots = %d\n"%(HE.numSlots()))
    f.write(" Encryption: t = %f\n"%(t_e))
    f.write(" SUM(%d): t = %f\n"%(f_sum, t_sum)) 
    f.write(" PROD(%d): t = %f\n"%(f_prod, t_prod)) 
    f.write(" SC.PROD(%d): t = %f\n"%(f_scpr, t_scpr)) 
    f.write(" LIMITS:\n")
    f.write("   - SUM: n = %d, t = %f t_avg = %f, td = %f, td_avg=%f\n"%(n_maxSum,
        t_maxSum, t_maxSum/n_maxSum, t_maxSd, t_maxSd/n_maxSum))
    f.write("   - PROD: n = %d, t = %f t_avg = %f, td = %f, td_avg=%f\n"%(n_maxMult,
        t_maxMult, t_maxMult/n_maxMult, t_maxMd, t_maxMd/n_maxMult))
    f.write("---------------------------------------------\n")
    f.close()

    return

# RUN THE ZILLION TESTS!
fN = 'ZillionTests1.txt'
i=0
for p,r in zip([3, 2, 2, 2, 2, 253, 257, 257, 257, 65537, 65537, 4294967311],
               [5, 16, 30, 40, 54, 5, 1, 2, 3, 4, 1, 2, 1]):
    for L in [10, 20, 30, 35, 40]:
        for d in [0,1]:
            i=i+1
            try:
                runTest(fN, i, p=p, r=r, L=L, d=d)
            except(ValueError):
                f = open(fN, 'a')
                f.write("\n TEST %d FAILED. \n"%(i))
                f.close()



# distutils: language = c++
#cython: language_level=3, boundscheck=False

"""PyCtxt. Ciphertext of Pyfhel, Python For Homomorphic Encryption Libraries.
"""
# -------------------------------- IMPORTS ------------------------------------
# Import Pyfhel and PyPtxt for operations
from .Pyfhel import Pyfhel
from .PyPtxt import PyPtxt

# Encoding types: 0-UNDEFINED, 1-INTEGER, 2-FRACTIONAL, 3-BATCH
from Pyfhel.util import ENCODING_t

# Dereferencing pointers in Cython in a secure way
from cython.operator cimport dereference as deref

# ----------------------------- IMPLEMENTATION --------------------------------
cdef class PyCtxt:
    """Ciphertext of Pyfhel. Contains a value/vector of encrypted ints/doubles.

    This class references SEAL, PALISADE and HElib ciphertexts, using the one 
    corresponding to the backend selected in Pyfhel. By default, it is SEAL.

    Attributes:
        * copy_ctxt (PyCtxt, optional): Other PyCtxt to deep copy.
        * pyfhel (Pyfhel, optional): Pyfhel instance needed to operate.
        * fileName (str|Path, optional): Load PyCtxt from this file.
                         Requires non-empty encoding.
        * encoding (str|type|int, optional): encoding type of the new PyCtxt.
    
    """
    def __cinit__(self,
                  PyCtxt copy_ctxt=None,
                  Pyfhel pyfhel=None,
                  fileName=None,
                  encoding=None):
        if (copy_ctxt): # If there is a PyCtxt to copy, override all arguments and copy
            self._ptr_ctxt = new Ciphertext(deref(copy_ctxt._ptr_ctxt))
            self._encoding = copy_ctxt._encoding
            if (copy_ctxt._pyfhel):
                self._pyfhel = copy_ctxt._pyfhel
        
        else:
            self._ptr_ctxt = new Ciphertext()
            if fileName:
                if not encoding:
                    raise TypeError("<Pyfhel ERROR> PyCtxt initialization with loading requires valid encoding")    
                self.from_file(fileName, encoding)
            else:
                self._encoding = to_ENCODING_t(encoding) if encoding else ENCODING_T.UNDEFINED
            if (pyfhel):
                self._pyfhel = pyfhel
            
    def __dealloc__(self):
        if self._ptr_ctxt != NULL:
            del self._ptr_ctxt
            
    @property
    def _encoding(self):
        """returns the encoding type"""
        return ENCODING_t(self._encoding)
    
    @_encoding.setter
    def _encoding(self, new_encoding):
        """Sets Encoding type: 0-UNDEFINED, 1-INTEGER, 2-FRACTIONAL, 3-BATCH""" 
        if not isinstance(new_encoding, ENCODING_t):
            raise TypeError("<Pyfhel ERROR> Encoding type of PyPtxt must be ENCODING_t")        
        self._encoding = new_encoding.value
        
    @_encoding.deleter
    def _encoding(self):
        """Sets Encoding to 1-UNDEFINED""" 
        self._encoding = ENCODING_t.UNDEFINED.value
        
    @property
    def _pyfhel(self):
        """A pyfhel instance, used for operations"""
        return self._pyfhel
    @_pyfhel.setter
    def _pyfhel(self, new_pyfhel):
        """Sets the pyfhel instance, used for operations""" 
        if not isinstance(new_pyfhel, Pyfhel):
            raise TypeError("<Pyfhel ERROR> new_pyfhel needs to be a Pyfhel class object")       
        self._pyfhel = new_pyfhel 
        
    cpdef int size_capacity(self) except +:
        """int: Maximum size the ciphertext can hold."""
        return self._ptr_ctxt.size_capacity()
     
    cpdef int size(self) except +:
        """int: Actual size of the ciphertext."""
        return self._ptr_ctxt.size()

    @property    
    def capacity(self):
        """int: Maximum size the ciphertext can hold."""
        return self._ptr_ctxt.size_capacity()

    @property
    def size(self):
        """int: Actual size of the ciphertext."""
        return self._ptr_ctxt.size()

    # =========================================================================
    # ================================== I/O ==================================
    # =========================================================================
    cpdef void to_file(self, fileName) except +:
        """to_file(Path fileName)
        
        Alias of `save` with input sanitizing.
        """
        self.save(_to_valid_file_str(fileName))

    cpdef void save(self, str fileName) except +:
        """save(str fileName)
        
        Save the ciphertext into a file. The file can new one or
        exist already, in which case it will be overwriten.

        Args:
            fileName: (:obj:`str`) File where the ciphertext will be stored.
        """
        cdef ofstream* outputter
        cdef string bFileName = fileName.encode('utf8')
        outputter = new ofstream(bFileName, binary)
        try:
            self._ptr_ctxt.save(deref(outputter))
        finally:
            del outputter

    cpdef bytes to_bytes(self) except +:
        """to_bytes()

        Serialize the ciphertext into a binary/bytes string.

        Return:
            * bytes: serialized ciphertext
        """
        cdef ostringstream outputter
        self._ptr_ctxt.save(outputter)
        return outputter.str()

    cpdef void from_file(self, fileName, encoding) except +:
        """from_file(str fileName)
        
        Alias of `load` with input sanitizer.
        """
        self.load(_to_valid_file_str(fileName, check=True), encoding)

    cpdef void load(self, str fileName, encoding) except +:
        """load(self, str fileName)
        
        Load the ciphertext from a file.

        Args:
            fileName: (:obj:`str`) File where the ciphertext is retrieved from.
            encoding: (:obj: `str`) String or type describing the encoding:
                'int' or int for IntegerEncoding (default),
                'float'/'fractional'/'double' or float for FractionalEncoding,
                'array'/'batch'/'matrix' or list for BatchEncoding

        """
        cdef ifstream* inputter
        cdef string bFileName = fileName.encode('utf8')
        inputter = new ifstream(bFileName,binary)
        try:
            self._ptr_ctxt.load(deref(inputter))
        finally:
            del inputter
        self._encoding = to_ENCODING_t(encoding).value

    cpdef void from_bytes(self, bytes content, encoding) except +:
        """from_bytes(bytes content)

        Recover the serialized ciphertext from a binary/bytes string.

        Args:
            content: (:obj:`bytes`) Python bytes object containing the PyCtxt.
            encoding: (:obj: `str`) String or type describing the encoding:
                'int' or int for IntegerEncoding (default),
                'float'/'fractional'/'double' or float for FractionalEncoding,
                'array'/'batch'/'matrix' or list for BatchEncoding
        """
        cdef stringstream inputter
        inputter.write(content,len(content))
        self._ptr_ctxt.load(inputter)
        self._encoding = to_ENCODING_t(encoding).value

            
    # =========================================================================
    # ============================= OPERATIONS ================================
    # =========================================================================          
    def __neg__(self):
        """Negates this ciphertext.
        """
        self._pyfhel.negate(self)
        
    def __add__(self, other):
        """Sums this ciphertext with either another PyCtx or a PyPtxt plaintext.
        
        Sums with a PyPtxt/PyCtxt, storing the result a new ciphertext.

        Args:
            other (PyCtxt|PyPtxt): Second summand.

        Returns:
            (PyCtxt): Ciphertext resulting of substraction

        Raise:
            TypeError: if other doesn't have a valid type.
        """
        if isinstance(other, PyCtxt):
            return self._pyfhel.add(self, other, in_new_ctxt=True)
        elif isinstance(other, PyPtxt):
            return self._pyfhel.add_plain(self, other, in_new_ctxt=True)
        elif isinstance(other, (int, float)):
            if self._encoding == ENCODING_t.INTEGER:
                other = self._pyfhel.encodeInt(int(other))
                return self._pyfhel.add_plain(self, other, in_new_ctxt=True)
            elif self._encoding == ENCODING_t.FRACTIONAL:
                other = self._pyfhel.encodeFrac(float(other))
                return self._pyfhel.add_plain(self, other, in_new_ctxt=True)
        else:
            raise TypeError("<Pyfhel ERROR> other summand must be either PyCtxt or PyPtxt")
    
    def __radd__(self, other): return self.__add__(other)
    def __iadd__(self, other):
        """Sums this ciphertext with either another PyCtx or a PyPtxt plaintext.
        
        Sums with a PyPtxt/PyCtxt, storing the result in this ciphertext.

        Args:
            other (PyCtxt|PyPtxt): Second summand.
            
        Raise:
            TypeError: if other doesn't have a valid type.
        """
        if isinstance(other, PyCtxt):
            self._pyfhel.add(self, other, in_new_ctxt=False)
        elif isinstance(other, PyPtxt):
            self._pyfhel.add_plain(self, other, in_new_ctxt=False)
        elif isinstance(other, (int, float)):
            if self._encoding == ENCODING_t.INTEGER:
                other = self._pyfhel.encodeInt(int(other))
                self._pyfhel.add_plain(self, other, in_new_ctxt=False)
            elif self._encoding == ENCODING_t.FRACTIONAL:
                other = self._pyfhel.encodeFrac(float(other))
                self._pyfhel.add_plain(self, other, in_new_ctxt=False)
        else:
            raise TypeError("<Pyfhel ERROR> other summand must be either PyCtxt or PyPtxt")
        return self


    def __sub__(self, other):
        """Substracts this ciphertext with either another PyCtxt or a PyPtxt plaintext.
        
        Substracts with a PyPtxt/PyCtxt, storing the result in a new ciphertext.

        Args:
            other (PyCtxt|PyPtxt): Substrahend, to be substracted from this ciphertext.
        Returns:
            (PyCtxt): Ciphertext resulting of substraction

        Raise:
            TypeError: if other doesn't have a valid type.
        """
        if isinstance(other, PyCtxt):
            return self._pyfhel.sub(self, other, in_new_ctxt=True)
        elif isinstance(other, PyPtxt):
            return self._pyfhel.sub_plain(self, other, in_new_ctxt=True)
        elif isinstance(other, (int, float)):
            if self._encoding == ENCODING_t.INTEGER:
                other = self._pyfhel.encodeInt(int(other))
                return self._pyfhel.sub_plain(self, other, in_new_ctxt=True)
            elif self._encoding == ENCODING_t.FRACTIONAL:
                other = self._pyfhel.encodeFrac(float(other))
                return self._pyfhel.sub_plain(self, other, in_new_ctxt=True)
        else:
            raise TypeError("<Pyfhel ERROR> substrahend must be either PyCtxt or PyPtxt")
    
    def __rsub__(self, other): return self.__sub__(other)
    def __isub__(self, other): 
        """Substracts this ciphertext with either another PyCtxt or a PyPtxt plaintext.
        
        Substracts with a PyPtxt/PyCtxt, storing the result in this ciphertext.

        Args:
            other (PyCtxt|PyPtxt): Substrahend, to be substracted from this ciphertext.
            
        Raise:
            TypeError: if other doesn't have a valid type.
        """
        if isinstance(other, PyCtxt):
            self._pyfhel.sub(self, other, in_new_ctxt=False)
        elif isinstance(other, PyPtxt):
            self._pyfhel.sub_plain(self, other, in_new_ctxt=False)
        elif isinstance(other, (int, float)):
            if self._encoding == ENCODING_t.INTEGER:
                other = self._pyfhel.encodeInt(int(other))
                self._pyfhel.sub_plain(self, other, in_new_ctxt=False)
            elif self._encoding == ENCODING_t.FRACTIONAL:
                other = self._pyfhel.encodeFrac(float(other))
                self._pyfhel.sub_plain(self, other, in_new_ctxt=False)
        else:
            raise TypeError("<Pyfhel ERROR> substrahend must be either PyCtxt or PyPtxt")
        return self
                        
    def __mul__(self, other):
        """Multiplies this ciphertext with either another PyCtxt or a PyPtxt plaintext.
        
        Multiplies with a PyPtxt/PyCtxt, storing the result in a new ciphertext.

        Args:
            other (PyCtxt|PyPtxt): Multiplier, to be multiplied with this ciphertext.

        Returns:
            (PyCtxt): Ciphertext resulting of multiplication

        Raise:
            TypeError: if other doesn't have a valid type.
        """
        if isinstance(other, PyCtxt):
            return self._pyfhel.multiply(self, other, in_new_ctxt=True)
        elif isinstance(other, PyPtxt):
            return self._pyfhel.multiply_plain(self, other, in_new_ctxt=True)
        elif isinstance(other, (int, float)):
            if self._encoding == ENCODING_t.INTEGER:
                other = self._pyfhel.encodeInt(int(other))
                return self._pyfhel.multiply_plain(self, other, in_new_ctxt=True)
            elif self._encoding == ENCODING_t.FRACTIONAL:
                other = self._pyfhel.encodeFrac(float(other))
                return self._pyfhel.multiply_plain(self, other, in_new_ctxt=True)
        else:
            raise TypeError("<Pyfhel ERROR> multiplicand must be either PyCtxt, PyPtxt or int|float"
                            "(is %s instead)"%(type(other)))
     
    def __rmul__(self, other): return self.__mul__(other)
    def __imul__(self, other): 
        """Multiplies this ciphertext with either another PyCtxt or a PyPtxt plaintext.
        
        Multiplies with a PyPtxt/PyCtxt, storing the result in this ciphertext.

        Args:
            other (PyCtxt|PyPtxt): Multiplier, to be multiplied with this ciphertext.

        Returns:
            (PyCtxt): Ciphertext resulting of multiplication

        Raise:
            TypeError: if other doesn't have a valid type.
        """
        if isinstance(other, PyCtxt):
            self._pyfhel.multiply(self, other, in_new_ctxt=False)
        elif isinstance(other, PyPtxt):
            self._pyfhel.multiply_plain(self, other, in_new_ctxt=False)
        elif isinstance(other, (int, float)):
            if self._encoding == ENCODING_t.INTEGER:
                other = self._pyfhel.encodeInt(int(other))
                self._pyfhel.multiply_plain(self, other, in_new_ctxt=False)
            elif self._encoding == ENCODING_t.FRACTIONAL:
                other = self._pyfhel.encodeFrac(float(other))
                self._pyfhel.multiply_plain(self, other, in_new_ctxt=False)
        else:
            raise TypeError("<Pyfhel ERROR> multiplicand must be either PyCtxt, PyPtxt or int|float"
                            "(is %s instead)"%(type(other)))
        return self


    def __truediv__(self, divisor):
        """Multiplies this ciphertext with the inverse of divisor.
        
        This operation can only be done with plaintexts. Division between 
        two Ciphertexts is not possible.

        For IntegerEncoding, the inverse is calculated as:
            inverse -> (divisor * inverse) mod p = 1

        For FractionalEncoding, the inverse is calculated as 1/divisor.

        Args:
            divisor (int|float|PyPtxt): divisor for the operation.
        """
        if isinstance(divisor, PyPtxt):
            divisor = divisor.decode()
        if not isinstance(divisor, (int, float)):
            raise TypeError("<Pyfhel ERROR> divisor must be float, int"
                            "or PyPtxt with those encodings (is %s:%s instead)"
                            %(str(divisor),type(divisor)))
        # Compute inverse. Int: https://stackoverflow.com/questions/4798654
        if self._encoding == ENCODING_t.INTEGER:
            divisor = int(divisor)
            p = self._pyfhel.getp()
            inverse = pow(divisor, p-2, p)
            inversePtxt = self._pyfhel.encodeInt(inverse)
        elif self._encoding == ENCODING_t.FRACTIONAL: # float. Standard inverse
            inverse = 1/float(divisor)
            inversePtxt = self._pyfhel.encodeFrac(inverse)
        else:
            raise TypeError("<Pyfhel ERROR> dividend encoding doesn't support"
                            "division (%s)"%(self._encoding))
        return self._pyfhel.multiply_plain(self, inversePtxt, in_new_ctxt=True)

    def __itruediv__(self, divisor):
        """Multiplies this ciphertext with the inverse of divisor.
        
        This operation can only be done with plaintexts. Division between 
        two Ciphertexts is not possible.

        For IntegerEncoding, the inverse is calculated as:
            inverse -> (divisor * inverse) mod p = 1

        For FractionalEncoding, the inverse is calculated as 1/divisor.

        Args:
            divisor (int|float|PyPtxt): divisor for the operation.
        """
        if isinstance(divisor, PyPtxt):
            divisor = divisor.decode()
        if not isinstance(divisor, (int, float)):
            raise TypeError("<Pyfhel ERROR> divisor must be float, int"
                            "or PyPtxt with those encodings (is %s:%s instead)"
                            %(str(divisor),type(divisor)))
        # Compute inverse. Int: https://stackoverflow.com/questions/4798654
        if self._encoding == ENCODING_t.INTEGER:
            divisor = int(divisor)
            p = self._pyfhel.getp()
            inverse = pow(divisor, p-2, p)
            inversePtxt = self._pyfhel.encodeInt(inverse)
        elif self._encoding == ENCODING_t.FRACTIONAL: # float. Standard inverse
            inverse = 1/float(divisor)
            inversePtxt = self._pyfhel.encodeFrac(inverse)
        else:
            raise TypeError("<Pyfhel ERROR> dividend encoding doesn't support"
                            "division (%s)"%(self._encoding))
        self._pyfhel.multiply_plain(self, inversePtxt, in_new_ctxt=False)
        return self

                                    
    def __pow__(self, exponent, modulo):
        """Exponentiates this ciphertext to the desired exponent.
        
        Exponentiates to the desired exponent.

        Args:
            exponent (int): Exponent for the power.
        """
        if(exponent==2):
            return self._pyfhel.square(self, in_new_ctxt=True)  
        else:
            return self._pyfhel.power(self, expon=exponent, in_new_ctxt=True)

    def __ipow__(self, exponent):
        """Exponentiates this ciphertext to the desired exponent, inplace.
        
        Exponentiates to the desired exponent.

        Args:
            exponent (int): Exponent for the power.
        """
        if(exponent==2):
            self._pyfhel.square(self, in_new_ctxt=False)  
        else:
            self._pyfhel.power(self, expon=exponent, in_new_ctxt=False)
        return self
                
    def __rshift__(self, k):
        """Rotates this ciphertext k positions to the right.
        Only works in batching mode.
        
        Args:
            k (int): Number of positions to rotate.
        """
        return self._pyfhel.rotate(self, -k, in_new_ctxt=True)

    def __irshift__(self, k):
        """Rotates this ciphertext k positions to the right, in-place.
        Only works in batching mode.

        Args:
            k (int): Number of positions to rotate.
        """
        self._pyfhel.rotate(self, -k, in_new_ctxt=False)
        return self

    def __lshift__(self, k):
        """Rotates this ciphertext k positions to the left.
        Only works in batching mode.

        Args:
            k (int): Number of positions to rotate.
        """
        return self._pyfhel.rotate(self, k, in_new_ctxt=True)

    def __ilshift__(self, k):
        """Rotates this ciphertext k positions to the left, in-place.
        Only works in batching mode.

        Args:
            k (int): Number of positions to rotate.
        """
        self._pyfhel.rotate(self, k, in_new_ctxt=False)
        return self

    def __invert__(self):
        """Relinarizes this ciphertext in-place.

        Requires valid relinearization keys with a bitcount higher than the
        current size of this ciphertext.
        """
        self._pyfhel.relinearize(self)
        return self

    # =========================================================================
    # ============================ ENCR/DECR/CMP ==============================
    # =========================================================================

    def __len__(self):
        return self.size()

    def __int__(self):
        if (self._encoding != ENCODING_T.INTEGER):
            raise RuntimeError("<Pyfhel ERROR> wrong PyCtxt encoding (not INTEGER)")
        return self._pyfhel.decryptInt(self)

    def __float__(self):
        if (self._encoding != ENCODING_T.FRACTIONAL):
            raise RuntimeError("<Pyfhel ERROR> wrong PyCtxt encoding (not FRACTIONAL)")
        return self._pyfhel.decryptFrac(self)
    
    def __repr__(self):
        return "<Pyfhel Ciphertext at {}, encoding={}, size={}/{}, noiseBudget={}>".format(
                hex(id(self)),
                ENCODING_t(self._encoding).name,
                self.size(),
                self.size_capacity(),
                self._pyfhel.noiseLevel(self) if self._pyfhel else "-")
                
    def __bytes__(self):
        return self.to_bytes()

    def encrypt(self, value):
        self._pyfhel.encrypt(value, self)
    
    def decrypt(self):
        return self._pyfhel.decrypt(self, decode_value=True)



# Fractional / float Demo for Pyfhel, covering the different ways of encrypting
#   and decrypting.

from Pyfhel import Pyfhel, PyPtxt, PyCtxt

print("==============================================================")
print("===================== Pyfhel FRACTIONAL ======================")
print("==============================================================")


print("1. Creating Context and KeyGen in a Pyfhel Object. Using 64 ")
print("     bits for integer part and 32 bits for decimal part.")
HE = Pyfhel()           # Creating empty Pyfhel object
HE.contextGen(p=65537, base=2, intDigits=64, fracDigits = 32) 
                        # Generating context. The value of p is important.
                        #  There are many configurable parameters on this step
                        #  More info in Demo_ContextParameters.py, and
                        #  in the docs of the function (link to docs in README)
HE.keyGen()             # Key Generation.
print(HE)

print("2. Encrypting Fractionals")
float1 = -7.3
float2 = 3.4
ctxt1 = HE.encryptFrac(float1) # Encryption makes use of the public key
ctxt2 = HE.encryptFrac(float2) # For integers, encryptInt function is used.
print("    int ",float1,'-> ctxt1 ', ctxt1)
print("    int ",float2,'-> ctxt2 ', ctxt2)



print("3. Operating with encrypted fractionals")
ctxtSum = ctxt1 + ctxt2         # `ctxt1 += ctxt2` for quicker inplace operation
ctxtSub = ctxt1 - ctxt2         # `ctxt1 -= ctxt2` for quicker inplace operation
ctxtMul = ctxt1 * ctxt2         # `ctxt1 *= ctxt2` for quicker inplace operation
print(ctxtSum)
print(ctxtSub)
print(ctxtMul)

print("4. Decrypting result:")
resSum = HE.decryptFrac(ctxtSum) # Decryption must use the corresponding function
                                #  decryptInt.
resSub = HE.decryptFrac(ctxtSub) 
resMul = HE.decryptFrac(ctxtMul)
print("     addition:       decrypt(ctxt1 + ctxt2) =  ", resSum)
print("     substraction:   decrypt(ctxt1 - ctxt2) =  ", resSub)
print("     multiplication: decrypt(ctxt1 + ctxt2) =  ", resMul)

print("NOTE: Very accurate! Let's try lowering the fracDigits from the context.")
print("==============================================================")

print("1bis. Creating Context and KeyGen in a Pyfhel Object. Using 64 ")
print("     bits for integer part and 3 bits for decimal part.")
HE = Pyfhel()           # Creating empty Pyfhel object
HE.contextGen(p=65537, base=2, intDigits=64, fracDigits = 3)
                        #  There are many configurable parameters on this step
                        #  More info in Demo_ContextParameters.py, and
                        #  in the docs of the function (link to docs in README)
HE.keyGen()             # Key Generation.

print("2bis. Encrypting Fractionals")
float1 = -7.3
float2 = 3.4
ctxt1 = HE.encryptFrac(float1) # Encryption makes use of the public key
ctxt2 = HE.encryptFrac(float2) # For integers, encryptInt function is used.
print("    int ",float1,'-> ctxt1 ', type(ctxt1))
print("    int ",float2,'-> ctxt2 ', type(ctxt2))



print("3bis. Operating with encrypted fractionals")
ctxtSum = ctxt1 + ctxt2         # `ctxt1 += ctxt2` for quicker inplace operation
ctxtSub = ctxt1 - ctxt2         # `ctxt1 -= ctxt2` for quicker inplace operation
ctxtMul = ctxt1 * ctxt2         # `ctxt1 *= ctxt2` for quicker inplace operation


print("4bis. Decrypting result:")
resSum = HE.decryptFrac(ctxtSum) # Decryption must use the corresponding function
                                #  decryptInt.
resSub = HE.decryptFrac(ctxtSub) 
resMul = HE.decryptFrac(ctxtMul)
print("     addition:       decrypt(ctxt1 + ctxt2) =  ", resSum)
print("     substraction:   decrypt(ctxt1 - ctxt2) =  ", resSub)
print("     multiplication: decrypt(ctxt1 + ctxt2) =  ", resMul)

print("NOTE: As you can see, the accuracy drops! As usual, there is a tradeoff.")



# Global modules
import os
import sys
import tempfile
import time
import unittest
from pathlib import Path
from io import StringIO
import contextlib
import numpy as np

# Local module
from Pyfhel import PyCtxt, Pyfhel, PyPtxt
from Pyfhel.util import ENCODING_t


# Redirect stdout of Demos to a string, recover if error.
# https://stackoverflow.com/a/41614610/9670056
@contextlib.contextmanager
def stdoutIO(stdout=None):
    old = sys.stdout
    if stdout is None:
        stdout = StringIO()
    sys.stdout = stdout
    yield stdout
    sys.stdout = old

# Demo execution from https://stackoverflow.com/a/41658338/9670056
def execfile(filepath, globals=None, locals=None):
    if globals is None:
        globals = {}
    globals.update({
        "__file__": filepath,
        "__name__": "__main__",
    })
    with open(filepath, 'rb') as file:
        with stdoutIO() as s:
            exec(compile(file.read(), filepath, 'exec'), globals, locals)

# Examples folder
EXAMPLES_FOLDER = Path('../examples')

# Value of p for batching: p=1964769281


class PyfhelTestCase(unittest.TestCase):
    def setUp(self):
        self.t0 = time.time()

    def tearDown(self):
        sys.stderr.write("({}s) ...".format(round(time.time() - self.t0, 3)))

    def test_PTXT_CTXT(self):
        pass

    def test_PyPtxt_creation_deletion(self):
        try:
            ptxt = PyPtxt()
            ptxt2 = PyPtxt(copy_ptxt=ptxt)
            pyfhel = Pyfhel()
            ptxt3 = PyPtxt(pyfhel=pyfhel)
            ptxt4 = PyPtxt(copy_ptxt=ptxt3)
        except Exception as err:
            self.fail("PyPtxt() creation failed unexpectedly: ", err)
        self.assertEqual(ptxt._encoding, ENCODING_t.UNDEFINED)
        ptxt._encoding = ENCODING_t.INTEGER
        self.assertEqual(ptxt._encoding, ENCODING_t.INTEGER)
        del (ptxt._encoding)
        self.assertEqual(ptxt._encoding, ENCODING_t.UNDEFINED)
        ptxt._pyfhel = pyfhel
        ptxt2._pyfhel = ptxt._pyfhel
        try:
            del (ptxt)
        except Exception as err:
            self.fail("PyPtxt() deletion failed unexpectedly: ", err)

    def test_PyCtxt_creation_deletion(self):
        try:
            ctxt = PyCtxt()
            ctxt2 = PyCtxt(copy_ctxt=ctxt)
            pyfhel = Pyfhel()
            ctxt3 = PyCtxt(pyfhel=pyfhel)
            ctxt4 = PyCtxt(copy_ctxt=ctxt3)
        except Exception as err:
            self.fail("PyCtxt() creation failed unexpectedly: ", err)
        self.assertEqual(ctxt.size(), 2)
        self.assertEqual(ctxt._encoding, ENCODING_t.UNDEFINED)
        ctxt._encoding = ENCODING_t.FRACTIONAL
        self.assertEqual(ctxt._encoding, ENCODING_t.FRACTIONAL)
        del (ctxt._encoding)
        self.assertEqual(ctxt._encoding, ENCODING_t.UNDEFINED)
        self.assertEqual(ctxt.size(), 2)
        ctxt._pyfhel = pyfhel
        ctxt2._pyfhel = ctxt._pyfhel
        try:
            del (ctxt)
        except Exception as err:
            self.fail("PyCtxt() deletion failed unexpectedly: ", err)

    def test_Pyfhel_1_GENERATION(self):
        pass

    def test_Pyfhel_1a_creation_deletion(self):
        try:
            pyfhel = Pyfhel()
        except Exception as err:
            self.fail("Pyfhel() creation failed unexpectedly: ", err)
        try:
            del (pyfhel)
        except Exception as err:
            self.fail("Pyfhel() deletion failed unexpectedly: ", err)

    def test_Pyfhel_1b_context_n_key_generation(self):
        pyfhel = Pyfhel()
        pyfhel.contextGen(65537)
        pyfhel.keyGen()

    def test_Pyfhel_1c_rotate_key_generation(self):
        pyfhel = Pyfhel()
        pyfhel.contextGen(65537)
        pyfhel.keyGen()
        pyfhel.rotateKeyGen(30)
        pyfhel.rotateKeyGen(1)
        pyfhel.rotateKeyGen(60)
        self.assertRaises(SystemError, lambda: pyfhel.rotateKeyGen(61))
        self.assertRaises(SystemError, lambda: pyfhel.rotateKeyGen(0))

    def test_Pyfhel_1d_relin_key_generation(self):
        pyfhel = Pyfhel()
        pyfhel.contextGen(65537)
        pyfhel.keyGen()
        pyfhel.relinKeyGen(30, 5)
        pyfhel.relinKeyGen(1, 5)
        pyfhel.relinKeyGen(60, 5)
        self.assertRaises(SystemError, lambda: pyfhel.relinKeyGen(61, 5))
        self.assertRaises(SystemError, lambda: pyfhel.relinKeyGen(0, 5))

    def test_Pyfhel_2_ENCODING(self):
        pass

    def test_Pyfhel_2a_encode_decode_int(self):
        pyfhel = Pyfhel()
        pyfhel.contextGen(p=65537)
        pyfhel.keyGen()
        ptxt = pyfhel.encodeInt(127)
        self.assertEqual(
            ptxt.to_poly_string(), b"1x^6 + 1x^5 + 1x^4 + 1x^3 + 1x^2 + 1x^1 + 1"
        )
        self.assertEqual(pyfhel.decodeInt(ptxt), 127)
        ptxt2 = PyPtxt(ptxt)
        pyfhel.encodeInt(-2, ptxt)
        self.assertEqual(ptxt.to_poly_string(), b"10000x^1")
        self.assertEqual(pyfhel.decodeInt(ptxt), -2)
        self.assertEqual(pyfhel.decodeInt(ptxt2), 127)

    def test_Pyfhel_2b_encode_decode_float(self):
        pyfhel = Pyfhel()
        pyfhel.contextGen(p=65537, m=8192, base=2, intDigits=80, fracDigits=20)
        pyfhel.keyGen()
        ptxt = pyfhel.encodeFrac(19.30)
        self.assertTrue(ptxt.to_poly_string(), b"9x^8190 + 1x^4 + 1x^1 + 1")
        self.assertEqual(round(pyfhel.decodeFrac(ptxt), 2), 19.30)
        pyfhel.encodeFrac(-2.25, ptxt)
        self.assertEqual(ptxt.to_poly_string(), b"1x^8190 + 10000x^1")
        self.assertEqual(round(pyfhel.decodeFrac(ptxt), 2), -2.25)

    def test_Pyfhel_2c_encode_decode_batch(self):
        pyfhel = Pyfhel()
        pyfhel.contextGen(p=1964769281, m=8192, base=2, sec=192, flagBatching=True)
        pyfhel.keyGen()
        self.assertTrue(pyfhel.batchEnabled())
        ptxt = pyfhel.encodeBatch([1, 2, 3, 4, 5, 6])
        self.assertEqual(pyfhel.getnSlots(), 8192)
        self.assertEqual(pyfhel.decodeBatch(ptxt)[:6], [1, 2, 3, 4, 5, 6])

    def test_Pyfhel_2d_encode_decode_array(self):
        pyfhel = Pyfhel()
        pyfhel.contextGen(p=1964769281, m=8192, base=2, sec=192, flagBatching=True)
        pyfhel.keyGen()
        self.assertTrue(pyfhel.batchEnabled())
        ptxt = pyfhel.encodeArray(np.array([1, 2, 3, 4, 5, 6], dtype=np.int64))
        self.assertEqual(pyfhel.getnSlots(), 8192)
        self.assertTrue(
            np.alltrue(pyfhel.decodeArray(ptxt)[:6] == np.array([1, 2, 3, 4, 5, 6], dtype=np.int64))
        )

    def test_Pyfhel_3_ENCRYPTING(self):
        pass

    def test_Pyfhel_3a_encrypt_decrypt_int(self):
        pyfhel = Pyfhel()
        pyfhel.contextGen(p=65537)
        pyfhel.keyGen()
        ctxt = pyfhel.encryptInt(127)
        self.assertEqual(pyfhel.decryptInt(ctxt), 127)
        ctxt2 = PyCtxt(ctxt)
        pyfhel.encryptInt(-2, ctxt)
        self.assertEqual(pyfhel.decryptInt(ctxt), -2)
        self.assertEqual(pyfhel.decryptInt(ctxt), -2)
        self.assertEqual(pyfhel.decryptInt(ctxt2), 127)

    def test_Pyfhel_3b_encrypt_decrypt_float(self):
        pyfhel = Pyfhel()
        pyfhel.contextGen(p=65537, m=8192, base=2, intDigits=80, fracDigits=20)
        pyfhel.keyGen()
        ctxt = pyfhel.encryptFrac(19.30)
        self.assertEqual(round(pyfhel.decryptFrac(ctxt), 2), 19.30)
        pyfhel.encryptFrac(-2.25, ctxt)
        self.assertEqual(round(pyfhel.decryptFrac(ctxt), 2), -2.25)

    def test_Pyfhel_3c_encrypt_decrypt_batch(self):
        pyfhel = Pyfhel()
        pyfhel.contextGen(p=1964769281, m=8192, base=2, sec=192, flagBatching=True)
        pyfhel.keyGen()
        self.assertTrue(pyfhel.batchEnabled())
        ctxt = pyfhel.encryptBatch([1, 2, 3, 4, 5, 6])
        self.assertEqual(pyfhel.getnSlots(), 8192)
        self.assertEqual(pyfhel.decryptBatch(ctxt)[:6], [1, 2, 3, 4, 5, 6])

        # print(self.ptxt.to_string())

    def test_Pyfhel_3d_encrypt_decrypt_array(self):
        pyfhel = Pyfhel()
        pyfhel.contextGen(p=1964769281, m=8192, base=2, sec=192, flagBatching=True)
        pyfhel.keyGen()
        self.assertTrue(pyfhel.batchEnabled())
        ctxt = pyfhel.encryptArray(np.array([1, 2, 3, 4, 5, 6], dtype=np.int64))
        self.assertEqual(pyfhel.getnSlots(), 8192)
        self.assertTrue(
            np.alltrue(pyfhel.decryptArray(ctxt)[:6] == np.array([1, 2, 3, 4, 5, 6], dtype=np.int64))
        )

    def test_Pyfhel_4_OPERATIONS(self):
        pass

    def test_Pyfhel_4a_operations_integer(self):
        pyfhel = Pyfhel()
        pyfhel.contextGen(p=1964769281, m=8192, base=3, sec=192)
        pyfhel.keyGen()

        ctxti = pyfhel.encryptInt(127)
        ctxti2 = pyfhel.encryptInt(-2)
        ptxti = pyfhel.encodeInt(3)

        ctxt_add = pyfhel.add(ctxti, ctxti2, in_new_ctxt=True)
        ctxt_add2 = pyfhel.add_plain(ctxti, ptxti, in_new_ctxt=True)
        ctxt_sub = pyfhel.sub(ctxti, ctxti2, in_new_ctxt=True)
        ctxt_sub2 = pyfhel.sub_plain(ctxti, ptxti, in_new_ctxt=True)
        ctxt_mult = pyfhel.multiply(ctxti, ctxti2, in_new_ctxt=True)
        ctxt_mult2 = pyfhel.multiply_plain(ctxti, ptxti, in_new_ctxt=True)

        self.assertEqual(pyfhel.decryptInt(ctxt_add), 125)
        self.assertEqual(pyfhel.decryptInt(ctxt_add2), 130)
        self.assertEqual(pyfhel.decryptInt(ctxt_sub), 129)
        self.assertEqual(pyfhel.decryptInt(ctxt_sub2), 124)
        self.assertEqual(pyfhel.decryptInt(ctxt_mult), -254)
        self.assertEqual(pyfhel.decryptInt(ctxt_mult2), 381)

    def test_Pyfhel_4b_operations_frac(self):
        pyfhel = Pyfhel()
        pyfhel.contextGen(p=1964769281, m=8192, base=3, sec=192)
        pyfhel.keyGen()

        ctxti = pyfhel.encryptFrac(19.37)
        ctxti2 = pyfhel.encryptFrac(-2.25)
        ptxti = pyfhel.encodeFrac(3.12)

        ctxt_add = pyfhel.add(ctxti, ctxti2, in_new_ctxt=True)
        ctxt_add2 = pyfhel.add_plain(ctxti, ptxti, in_new_ctxt=True)
        ctxt_sub = pyfhel.sub(ctxti, ctxti2, in_new_ctxt=True)
        ctxt_sub2 = pyfhel.sub_plain(ctxti, ptxti, in_new_ctxt=True)
        ctxt_mult = pyfhel.multiply(ctxti, ctxti2, in_new_ctxt=True)
        ctxt_mult2 = pyfhel.multiply_plain(ctxti, ptxti, in_new_ctxt=True)

        self.assertEqual(round(pyfhel.decryptFrac(ctxt_add), 2), 17.12)
        self.assertEqual(round(pyfhel.decryptFrac(ctxt_add2), 2), 22.49)
        self.assertEqual(round(pyfhel.decryptFrac(ctxt_sub), 2), 21.62)
        self.assertEqual(round(pyfhel.decryptFrac(ctxt_sub2), 2), 16.25)
        self.assertEqual(round(pyfhel.decryptFrac(ctxt_mult), 2), -43.58)
        self.assertEqual(round(pyfhel.decryptFrac(ctxt_mult2), 2), 60.43)

    def test_Pyfhel_4c_operations_batch_array(self):
        pyfhel = Pyfhel()
        pyfhel.contextGen(p=1964769281, m=8192, base=2, sec=192, flagBatching=True)
        pyfhel.keyGen()
        pyfhel.rotateKeyGen(60)
        ctxti = pyfhel.encryptBatch([1, 2, 3, 4, 5, 6])
        ctxti2 = pyfhel.encryptArray(np.array([-6, -5, -4, -3, -2, -1], dtype=np.int64))
        ptxti = pyfhel.encodeArray(np.array([12, 15, 18, 21, 24, 27], dtype=np.int64))

        ctxt_add = pyfhel.add(ctxti, ctxti2, in_new_ctxt=True)
        ctxt_add2 = pyfhel.add_plain(ctxti, ptxti, in_new_ctxt=True)
        ctxt_sub = pyfhel.sub(ctxti, ctxti2, in_new_ctxt=True)
        ctxt_sub2 = pyfhel.sub_plain(ctxti, ptxti, in_new_ctxt=True)
        ctxt_mult = pyfhel.multiply(ctxti, ctxti2, in_new_ctxt=True)
        ctxt_mult2 = pyfhel.multiply_plain(ctxti, ptxti, in_new_ctxt=True)
        ctxt_rotate = pyfhel.rotate(ctxti, -2, in_new_ctxt=True)
        ctxt_rotate2 = pyfhel.rotate(ctxti, 2, in_new_ctxt=True)

        self.assertEqual(pyfhel.decryptBatch(ctxt_add)[:6], [-5, -3, -1, 1, 3, 5])
        self.assertEqual(pyfhel.decryptBatch(ctxt_add2)[:6], [13, 17, 21, 25, 29, 33])
        self.assertEqual(pyfhel.decryptBatch(ctxt_sub)[:6], [7, 7, 7, 7, 7, 7])
        self.assertEqual(
            pyfhel.decryptBatch(ctxt_sub2)[:6], [-11, -13, -15, -17, -19, -21]
        )
        self.assertEqual(
            pyfhel.decryptBatch(ctxt_mult)[:6], [-6, -10, -12, -12, -10, -6]
        )
        self.assertEqual(
            pyfhel.decryptBatch(ctxt_mult2)[:6], [12, 30, 54, 84, 120, 162]
        )
        self.assertEqual(pyfhel.decryptBatch(ctxt_rotate)[:6], [0, 0, 1, 2, 3, 4])
        self.assertEqual(pyfhel.decryptBatch(ctxt_rotate2)[:6], [3, 4, 5, 6, 0, 0])

    def test_Pyfhel_5_IO_SAVE_RESTORE(self):
        pass

    def test_Pyfhel_5a_save_objects(self):
        pyfhel = Pyfhel()
        pyfhel.contextGen(p=1964769281, m=8192, base=2, sec=192, flagBatching=True)
        pyfhel.keyGen()
        pyfhel.rotateKeyGen(60)
        pyfhel.relinKeyGen(60,3)

        self.assertTrue(pyfhel.saveContext("context.pycon"))
        self.assertTrue(pyfhel.savepublicKey("public_k.pypk"))
        self.assertTrue(pyfhel.savesecretKey("secret_k.pysk"))
        self.assertTrue(pyfhel.saverelinKey("relin_k.pyrlk"))
        self.assertTrue(pyfhel.saverotateKey("rotate_k.pyrok"))

    def test_Pyfhel_5b_restore_objects(self):
        pyfhel = Pyfhel()
        self.assertTrue(pyfhel.restoreContext("context.pycon"))
        self.assertTrue(pyfhel.restoresecretKey("secret_k.pysk"))
        self.assertTrue(pyfhel.restorepublicKey("public_k.pypk"))
        self.assertTrue(pyfhel.restorerelinKey("relin_k.pyrlk"))
        self.assertTrue(pyfhel.restorerotateKey("rotate_k.pyrok"))
        os.remove("context.pycon")
        os.remove("secret_k.pysk")
        os.remove("public_k.pypk")
        os.remove("rotate_k.pyrok")

    def test_Pyfhel_5c_save_restore_all(self):
        pyfhel = Pyfhel()
        pyfhel.contextGen(p=1964769281, m=8192, base=2, sec=192, flagBatching=True)
        pyfhel.keyGen()
        pyfhel.rotateKeyGen(60)
        pyfhel.relinKeyGen(60, 4)
        # save all keys into temporary directory
        tmp_dir = tempfile.TemporaryDirectory()
        pyfhel.saveContext(tmp_dir.name + "/context")
        pyfhel.savepublicKey(tmp_dir.name + "/pub.key")
        pyfhel.savesecretKey(tmp_dir.name + "/sec.key")
        pyfhel.saverelinKey(tmp_dir.name + "/relin.key")
        pyfhel.saverotateKey(tmp_dir.name + "/rotate.key")
        # restore all keys
        pyfhel2 = Pyfhel()
        pyfhel.contextGen(p=1964769281, m=8192, base=2, sec=192, flagBatching=True)
        pyfhel2.restoreContext(tmp_dir.name + "/context")
        pyfhel2.restorepublicKey(tmp_dir.name + "/pub.key")
        pyfhel2.restoresecretKey(tmp_dir.name + "/sec.key")
        pyfhel2.restorerelinKey(tmp_dir.name + "/relin.key")
        pyfhel2.restorerotateKey(tmp_dir.name + "/rotate.key")

        # test encryption decryption
        ctxt1 = pyfhel.encryptBatch([42])
        self.assertEqual(
            pyfhel2.decryptBatch(ctxt1)[0],
            42,
            "decrypting with restored keys should work",
        )
        try:
            pyfhel2.rotate(ctxt1, -1)
            self.assertEqual(
                pyfhel2.decryptBatch(ctxt1)[1],
                42,
                "decrypting with restored keys should work",
            )
        except Exception as err:
            self.fail("PyPtxt() creation failed unexpectedly: ", err)

        # test ciphertext storing
        ctxt2 = pyfhel.encryptInt(42)
        ctxt2.save(tmp_dir.name + "/ctxt2")

        ctxt_restored = PyCtxt()
        ctxt_restored.load(tmp_dir.name + "/ctxt2", "int")
        self.assertEqual(
            pyfhel2.decryptInt(ctxt_restored), 42, "decrypting ciphertext should work"
        )
        tmp_dir.cleanup()

    def test_Pyfhel_5d_save_restore_int(self):
        pyfhel = Pyfhel()
        pyfhel.contextGen(p=1964769281, m=8192, base=2, sec=192, flagBatching=True)
        pyfhel.keyGen()
        pyfhel.rotateKeyGen(60)
        pyfhel.relinKeyGen(60, 4)
        # encrypt something
        ctxt = pyfhel.encryptInt(42)
        # save to temporary file
        tmp_dir = tempfile.TemporaryDirectory()
        tmp_file = os.path.join(tmp_dir.name, "ctxt")
        ctxt.save(tmp_file)
        # load from temporary file
        loaded = PyCtxt()
        loaded.load(tmp_file, "int")
        self.assertEqual(pyfhel.decryptInt(loaded), 42)
        tmp_dir.cleanup()

    def test_Pyfhel_5e_save_restore_float(self):
        pyfhel = Pyfhel()
        pyfhel.contextGen(p=1964769281, m=8192, base=2, sec=192, flagBatching=True)
        pyfhel.keyGen()
        pyfhel.rotateKeyGen(60)
        pyfhel.relinKeyGen(60, 4)
        # encrypt something
        ctxt = pyfhel.encryptFrac(3.125)
        # save to temporary file
        tmp_dir = tempfile.TemporaryDirectory()
        tmp_file = os.path.join(tmp_dir.name, "ctxt")
        ctxt.save(tmp_file)
        # load from temporary file
        loaded = PyCtxt()
        loaded.load(tmp_file, "float")
        self.assertEqual(pyfhel.decryptFrac(loaded), 3.125)
        tmp_dir.cleanup()

    def test_Pyfhel_5f_save_restore_batch(self):
        pyfhel = Pyfhel()
        pyfhel.contextGen(p=1964769281, m=8192, base=2, sec=192, flagBatching=True)
        pyfhel.keyGen()
        pyfhel.rotateKeyGen(60)
        pyfhel.relinKeyGen(60, 4)
        # encrypt something
        ctxt = pyfhel.encryptBatch([1, 2, 3, 4])
        # save to temporary file
        tmp_dir = tempfile.TemporaryDirectory()
        tmp_file = os.path.join(tmp_dir.name, "ctxt")
        ctxt.save(tmp_file)
        # load from temporary file
        loaded = PyCtxt()
        loaded.load(tmp_file, "batch")
        self.assertEqual(pyfhel.decryptBatch(loaded)[:4], [1, 2, 3, 4])
        tmp_dir.cleanup()

    def test_Pyfhel_6_DEMOS(self):
        pass

    def test_Pyfhel_6a_demo_helloworld(self):
        execfile(EXAMPLES_FOLDER / 'Demo_HelloWorld.py')
    
    def test_Pyfhel_6b_demo_float(self):
        execfile(EXAMPLES_FOLDER / 'Demo_Float.py')

    def test_Pyfhel_6c_demo_encoding(self):
        execfile(EXAMPLES_FOLDER / 'Demo_Encoding.py')

    def test_Pyfhel_6d_demo_encrypting(self):
        execfile(EXAMPLES_FOLDER / 'Demo_Encrypting.py')

    def test_Pyfhel_6e_demo_numpy(self):
        execfile(EXAMPLES_FOLDER / 'Demo_Numpy.py')

    def test_Pyfhel_6f_demo_SaveNRestore(self):
        execfile(EXAMPLES_FOLDER / 'Demo_SaveNRestore.py')

    def test_Pyfhel_6g_demo_simd(self):
        execfile(EXAMPLES_FOLDER / 'Demo_SIMD.py')

    def test_Pyfhel_6h_demo_ClientServer(self):
        execfile(EXAMPLES_FOLDER / 'Demo_ClientServer.py')

    def test_Pyfhel_6i_demo_MultDepth_n_relin(self):
        execfile(EXAMPLES_FOLDER / 'Demo_MultDepth_n_relin.py')
        
    def test_Pyfhel_6j_demo_MultDepth_n_relin(self):
        execfile(EXAMPLES_FOLDER / 'Demo_ContextParameters.py')

if __name__ == "__main__":
    unittest.main(verbosity=2)



# SIMD Demo for Pyfhel, showing the use of the library to operate 
# with multiple integers at once with the batching technique.


# This example uses sympy for primality testing, but only if sympy is installed
def find_lib(lib=""):
    import importlib
    lib_spec = importlib.util.find_spec(lib)
    return lib_spec is not None
sympy_installed = find_lib("sympy")
if sympy_installed:
    from sympy import isprime

from Pyfhel import Pyfhel, PyPtxt, PyCtxt
# Pyfhel class contains most of the functions.
# PyPtxt is the plaintext class
# PyCtxt is the ciphertext class

import numpy as np

print("==============================================================")
print("======================= Pyfhel BATCHING ======================")
print("==============================================================")


print("1. Creating Context and KeyGen in a Pyfhel Object. Careful about the context parameters!")
print("- p needs to be prime and p-1 must be multiple of 2*m.")
print("- m will determine the number of integer slots inside a single ciphertext")
HE = Pyfhel()           # Creating empty Pyfhel object
p = 1964769281
m = 8192
assert((p-1)%(2*m) == 0)
if (sympy_installed):
    assert (isprime(p))
HE.contextGen(p=p,m=m, base=2, sec=192, flagBatching=True)
HE.keyGen()             # Key Generation.
print("   We use Pyfhel.batchEnabled() to check if the current context allows batching, and getnSlots() to get the total number of integers that fit in a single ciphertext row (should be equal to m)")

print("2. Encrypting lists of Integers / NumPy arrays of size up to m. Empty slots are set to 0.")
list1 = [1,2,3,4,5,6]
list2 = [-2, -4, -6, -8, 0, 13]
ctxt1 = HE.encryptBatch(list1) # Encryption makes use of the public key
ctxt2 = HE.encryptBatch(list2) # For integers, encryptInt function is used.
print("    list1 ",list1,'-> ctxt1 ', type(ctxt1), str(ctxt1))
print("    list2 ",list2,'-> ctxt2 ', type(ctxt2), str(ctxt2))
arr1 = np.array([5,6,7,8], dtype=np.int64)
arr2 = np.array([-9, 10, -11, 12], dtype=np.int64)
ctxt3 = HE.encryptArray(arr1) 
ctxt4 = HE.encryptArray(arr2)

print("    arr1 ",arr1,'-> ctxt3 ', type(ctxt3), str(ctxt3))
print("    arr2 ",arr2,'-> ctxt4 ', type(ctxt4), str(ctxt4))


print("3. Operating in a SIMD manner with encrypted integers")
ctxtSum = ctxt1 + ctxt2
ctxtSub = ctxt3 - ctxt4
ctxtMul = ctxt1 * ctxt2
ctxtSum2 = ctxt3 + ctxt4
ctxtSub2 = ctxt1 - ctxt2      
ctxtMul2 = ctxt3 * ctxt2       

print("4. Decrypting results:")
resSum = HE.decryptBatch(ctxtSum) # Decryption must use the corresponding function
                                #  decryptInt.
resSub = HE.decryptBatch(ctxtSub) 
resMul = HE.decryptBatch(ctxtMul)
print("     addition:       decrypt(ctxt1 + ctxt2) =  ", resSum[:6])
print("     substraction:   decrypt(ctxt1 - ctxt2) =  ", resSub[:6])
print("     multiplication: decrypt(ctxt1 + ctxt2) =  ", resMul[:6])

resSum2 = np.array(HE.decryptBatch(ctxtSum2))
resSub2 = np.array(HE.decryptBatch(ctxtSub2))
resMul2 = np.array(HE.decryptBatch(ctxtMul2))
print("     addition:       decrypt(ctxt1 + ctxt3) =  ", resSum2[:4])
print("     substraction:   decrypt(ctxt3 - ctxt4) =  ", resSub2[:4])
print("     multiplication: decrypt(ctxt3 * ctxt2) =  ", resMul2[:4])



print("5. Trying out rotation:")
HE.rotateKeyGen(60)
print("     Initialized rotation keys: HE.rotateKeyGen(10)")
ctxt1_rot_r4 = ctxt1 >> 4   # Inplace equivalent: ctxt1 >>= 4 
ctxt2_rot_l2 = ctxt2 << 2   # Inplace equivalent: ctxt2 <<= 2

resrotr4 = np.array(HE.decryptBatch(ctxt1_rot_r4), dtype=np.int64)
resrotl2 = np.array(HE.decryptBatch(ctxt2_rot_l2), dtype=np.int64)

print("     ctxt1_rot_r4 = ctxt1 >> 4:  First 10 values:", resrotr4[:10])
print("                                 last 10 values:", resrotr4[-10:])
print("     ctxt2_rot_l2 = ctxt2 << 2:  First 10 values:", resrotl2[:10])
print("                                 last 10 values:", resrotl2[m//2-10:m//2])


from Pyfhel import Pyfhel, PyPtxt, PyCtxt
import tempfile
from pathlib import Path

# Using a temporary dir as a "secure channel"
# This can be changed into real communication using other python libraries.
secure_channel = tempfile.TemporaryDirectory()
sec_con = Path(secure_channel.name)
pk_file = sec_con / "mypk.pk"
contx_file = sec_con / "mycontx.con"


##### CLIENT
#HE Object Creation, including the public and private keys
HE = Pyfhel()    
HE.contextGen(p=65537, m=2**12) 
HE.keyGen() # Generates both a public and a private key

# Saving only the public key and the context
HE.savepublicKey(pk_file)
HE.saveContext(contx_file)

# Serializing two float values
a = 1.5
b = 2.5
ca = HE.encryptFrac(a)
cb = HE.encryptFrac(b)

ca.to_file(sec_con / "ca.ctxt")
cb.to_file(sec_con / "cb.ctxt")




##### SEMI-HONEST CLOUD
# Generating a second HE, acting as the honest-but-curious Cloud provider,
#  that will perform the operations and try to decrypt everything
HE_Cl = Pyfhel()    
HE_Cl.restoreContext(contx_file)
HE_Cl.restorepublicKey(pk_file)

# loading the two ciphertexts. There is clearly potential for improvement here
c2a = PyCtxt(pyfhel=HE_Cl, fileName=sec_con / "ca.ctxt", encoding=float)
c2b = PyCtxt(pyfhel=HE_Cl, fileName=sec_con / "cb.ctxt", encoding=float)

# Attempting to decrypt results raises an error (missing secret key)
#> ---------------------------------------------------------------------------
#> RuntimeError                              Traceback (most recent call last)
#> Pyfhel/Pyfhel.pyx in Pyfhel.Pyfhel.Pyfhel.decryptFrac()
#> RuntimeError: Missing a Private Key [...]
try:
    print(HE_Cl.decrypt(c2a))
    raise Exception("This should not be reached!")
except RuntimeError:
    print("The cloud tried to decrypt, but couldn't!")

# The cloud operates with the ciphertexts:
c_mean = (c2a + c2b) / 2

# And sends the result back
c_mean.to_file(sec_con / "c_mean.ctxt")




##### CLIENT
# Load and decrypt Result
c_res = PyCtxt(pyfhel=HE, fileName=sec_con / "c_mean.ctxt", encoding=float)
print(c_res.decrypt())


# Cleaning up secure channel
secure_channel.cleanup()



# =============================== SETUP.PY =====================================
# This file installs Pyfhel in your Python3 distribution. Use one of the two:
#   > python3 setup.py install
#   > python3 -m pip install .
# PYPI -> https://packaging.python.org/tutorials/packaging-projects/
#   > python3 setup.py sdist
#   > twine upload dist/*
#   > python3 setup.py clean --all

import shutil, glob, fileinput, re, os, sys, sysconfig, platform
from pathlib import Path

# Check that Python version is 3.5+
v_maj, v_min = sys.version_info[:2]
assert (v_maj, v_min) >= (3,5),\
    "Pyfhel requires Python 3.5+ (your version is {}.{}).".format(v_maj, v_min)



# -------------------------------- OPTIONS -------------------------------------
# Compile cython files (.pyx) into C++ (.cpp) files to ship with the library.
CYTHONIZE= False
if "--CYTHONIZE" in sys.argv:
    CYTHONIZE= True
    del sys.argv[sys.argv.index("--CYTHONIZE")]


# --------------------------- REQUIREMENT IMPORTS ------------------------------
# Create Extension modules written in C for Python
from setuptools import setup, Extension, find_packages

# Get directories for includes of both Python and Numpy
from distutils.sysconfig import get_python_inc

# Get requirements
with open("requirements.txt") as f:
    requirements = [req for req in f.read().split('\n') if req]
    
# --------------------------------- VERSION ------------------------------------
# Reading version info from Readme and hardcoding it in the __init__.py file
v_readme_regex = r'\[\_*v([0-9]+\.[0-9]+\.[0-9a-z]+)\_*\]'
with open('README.md') as readme:
    VERSION = re.findall(v_readme_regex,readme.read())[0]

v_init_regex = r'\"([0-9]+\.[0-9]+\.[0-9a-z]+)\"'
with open('Pyfhel/__init__.py') as f:
    s = f.read()
with open('Pyfhel/__init__.py', 'w') as f:
    f.write(re.sub(v_init_regex, '"'+VERSION+'"', s))

# Including Readme in the module as long description.
with open("README.md", "r") as fh:
    long_description = fh.read()


# ---------------------------- COMPILATION CONFIG ------------------------------
PYFHEL_PATH = Path('Pyfhel')
AFHEL_PATH = PYFHEL_PATH / 'Afhel'
SEAL_PATH = PYFHEL_PATH / 'SEAL' / 'SEAL' / 'seal'

# Scan a directory searching for all C++ files
def scan_cpp(dir, files=[]):
    for file in os.listdir(dir):
        path = os.path.join(dir, file)
        if os.path.isfile(path) and path.endswith(".cpp"):
            files.append(str(path))
    return files

# List all the .cpp files
local_sources = scan_cpp(SEAL_PATH,[str(AFHEL_PATH / 'Afseal.cpp')])

# Compile arguments for extensions
language            = "c++"
include_dirs        = [get_python_inc(),
                       str(PYFHEL_PATH), str(AFHEL_PATH), str(SEAL_PATH)]
define_macros=[("NPY_NO_DEPRECATED_API", "NPY_1_7_API_VERSION")]
extra_compile_flags = ["-DHAVE_CONFIG_H"]
if platform.system() == 'Windows':
    # Windows' MSVC2019 compiler doesn't have an O3 optimization
    #>https://docs.microsoft.com/en-us/cpp/build/reference/o-options-optimize-code
    extra_compile_flags += ["-O2"]
elif platform.system() == 'Darwin': # MacOS
    extra_compile_flags += ["-std=c++17","-O3","-mmacosx-version-min=10.12"]
else:  # Linux, GCC
    extra_compile_flags += ["-std=c++17","-O3"]


# --------------------------- LIBRARY COMPILATION ------------------------------
# Here we compile Afhel (with the backends) and bundle it into a static library
# Dynamic lybraries are much more complex to manage. In case this was necessary:
#> https://github.com/realead/commonso/blob/master/setup.py

# cpplibraries:
cpplibraries = ('Afhel', {'sources': local_sources,
                          'include_dirs':include_dirs,
                          'cflags':extra_compile_flags,
                          'macros':define_macros,})


# -------------------------------- EXTENSIONS ---------------------------------
ext = ".pyx" if CYTHONIZE else ".cpp"
ext_modules = [
         Extension(
             name="Pyfhel.Pyfhel",
             sources=[str(PYFHEL_PATH/("Pyfhel"+ext))],
             include_dirs=include_dirs,
             define_macros=define_macros,
             language=language,
             extra_compile_args=extra_compile_flags,
         ),
         Extension(
             name="Pyfhel.PyPtxt",
             sources=[str(PYFHEL_PATH/("PyPtxt"+ext))],
             include_dirs=include_dirs,
             define_macros=define_macros,
             language=language,
             extra_compile_args=extra_compile_flags,
         ),
         Extension(
             name="Pyfhel.PyCtxt",
             sources=[str(PYFHEL_PATH/("PyCtxt"+ext))],
             include_dirs=include_dirs,
             define_macros=define_macros,
             language=language,
             extra_compile_args=extra_compile_flags,
         ),   
]
if CYTHONIZE:
    from Cython.Build import cythonize
    ext_modules=cythonize(ext_modules)

# --------------------------------- CLEANER -----------------------------------
# Tired of cleaning all compilation and distribution by hand
from distutils.cmd import Command
class FlushCommand(Command):
    """Custom clean command to tidy up the project root."""
    CLEAN_FILES = './build ./dist ./*.pyc ./*.tgz ./*.egg-info'.split(' ')
    user_options = []
    def initialize_options(self):   pass
    def finalize_options(self):     pass
    def run(self):
        here = os.getcwd()
        for path_spec in self.CLEAN_FILES:
            # Make paths absolute and relative to this path
            abs_paths = glob.glob(os.path.normpath(os.path.join(here, path_spec)))
            for path in [str(p) for p in abs_paths]:
                if not path.startswith(here):
                    # Die if path in CLEAN_FILES is absolute + outside this directory
                    raise ValueError("%s is not a path inside %s" % (path, here))
                print('removing %s' % os.path.relpath(path))
                shutil.rmtree(path)

                
# ---------------------------------- NUMPY ------------------------------------
# We need to know the headers of numpy for compilation. For this, we use
#  our own build_ext function (https://stackoverflow.com/questions/54117786)
def my_build_ext(pars):
    # import delayed:
    from setuptools.command.build_ext import build_ext as _build_ext
    # include_dirs adjusted: 
    class build_ext(_build_ext):
        def finalize_options(self):
            _build_ext.finalize_options(self)
            # Prevent numpy from thinking it is still in its setup process:
            # __builtins__.__NUMPY_SETUP__ = False
            import numpy
            self.include_dirs.append(numpy.get_include())
    #object returned:
    return build_ext(pars)

# -------------------------------- INSTALLER ----------------------------------
setup(
    name            = "Pyfhel",
    version         = VERSION,
    author          = "Alberto Ibarrondo",
    author_email    = "ibarrond@eurecom.fr",
    description     = "Python for Homomorphic Encryption Libraries",
    long_description= long_description,
    long_description_content_type="text/markdown",
    keywords        = "homomorphic encryption cython cryptography",
    license         = "GNU GPLv3",
    url             = "https://github.com/ibarrond/Pyfhel",     
    setup_requires  =["setuptools>=45.0",
                      "numpy>=1.16.0"],
    install_requires=requirements,
    classifiers     =[
        "Programming Language :: C++",
        "Programming Language :: Cython",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: Implementation :: CPython",
        "Development Status :: 5 - Production/Stable",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: Unix",
        "Operating System :: POSIX",
        "Operating System :: Microsoft :: Windows",
        "Topic :: Security",
        "Topic :: Security :: Cryptography",
    ],
    zip_safe=False,
    packages=find_packages(),
    ext_modules=ext_modules,  
    test_suite=str(PYFHEL_PATH / "test.py"),
    libraries=[cpplibraries],
    cmdclass={'flush': FlushCommand,
              'build_ext' : my_build_ext},
)



from .Pyfhel import Pyfhel
from .PyCtxt import PyCtxt
from .PyPtxt import PyPtxt

__all__    = ["Pyfhel", "PyCtxt", "PyPtxt"]
__name__   = "Pyfhel"
__author__ = "Alberto Ibarrondo"

__version__ = "2.2.3"
